# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Pylint Code Analysis Plugin.
"""

# Standard library imports
from typing import List

# Third party imports
from qtpy.QtCore import Qt, Signal, Slot

# Local imports
from spyder.api.exceptions import SpyderAPIError
from spyder.api.plugins import Plugins, SpyderDockablePlugin
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.translations import _
from spyder.plugins.editor.api.run import FileRun
from spyder.plugins.mainmenu.api import ApplicationMenus, SourceMenuSections
from spyder.plugins.pylint.confpage import PylintConfigPage
from spyder.plugins.pylint.main_widget import PylintWidget
from spyder.plugins.run.api import (
    RunContext, RunConfiguration, PossibleRunResult, run_execute,
    ExtendedRunExecutionParameters, RunExecutor)


class PylintActions:
    AnalyzeCurrentFile = 'run analysis'


class Pylint(SpyderDockablePlugin, RunExecutor):

    NAME = "pylint"
    WIDGET_CLASS = PylintWidget
    CONF_SECTION = NAME
    CONF_WIDGET_CLASS = PylintConfigPage
    REQUIRES = [Plugins.Preferences, Plugins.Editor, Plugins.Run]
    OPTIONAL = [Plugins.MainMenu, Plugins.Projects]
    TABIFY = [Plugins.VariableExplorer, Plugins.Help]
    CONF_FILE = False
    DISABLE_ACTIONS_WHEN_HIDDEN = False

    # --- Signals
    sig_edit_goto_requested = Signal(str, int, str)
    """
    This signal will request to open a file in a given row and column
    using a code editor.

    Parameters
    ----------
    path: str
        Path to file.
    row: int
        Cursor starting row position.
    word: str
        Word to select on given row.
    """

    # ---- SpyderPluginV2 API
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _("Code Analysis")

    @staticmethod
    def get_description():
        return _("Analyze code and view the results from both static and "
                 "real-time analysis.")

    @classmethod
    def get_icon(cls):
        return cls.create_icon("pylint")

    def on_initialize(self):
        widget = self.get_widget()

        # Expose widget signals at the plugin level
        widget.sig_edit_goto_requested.connect(self.sig_edit_goto_requested)
        widget.sig_start_analysis_requested.connect(self.start_code_analysis)

        # To have a reference to the run action of this plugin
        self.run_action = None

        # Run configuration
        self.executor_configuration = [
            {
                'input_extension': 'py',
                'context': {'name': 'File'},
                'output_formats': [],
                'configuration_widget': None,
                'requires_cwd': False,
                'priority': 4
            },
        ]

    @on_plugin_available(plugin=Plugins.Editor)
    def on_editor_available(self):
        widget = self.get_widget()
        editor = self.get_plugin(Plugins.Editor)

        # Connect to Editor
        widget.sig_edit_goto_requested.connect(editor.load)
        editor.sig_editor_focus_changed.connect(self._set_filename)

    @on_plugin_available(plugin=Plugins.Preferences)
    def on_preferences_available(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.register_plugin_preferences(self)

    @on_plugin_available(plugin=Plugins.Projects)
    def on_projects_available(self):
        # Connect to projects
        projects = self.get_plugin(Plugins.Projects)

        projects.sig_project_loaded.connect(self._set_project_dir)
        projects.sig_project_closed.connect(self._unset_project_dir)

    @on_plugin_available(plugin=Plugins.Run)
    def on_run_available(self):
        run = self.get_plugin(Plugins.Run)
        run.register_executor_configuration(self, self.executor_configuration)

        self.run_action = run.create_run_in_executor_button(
            RunContext.File,
            self.NAME,
            text=_("Run code analysis"),
            tip=_("Run code analysis"),
            icon=self.create_icon("pylint"),
            shortcut_context='pylint',
            register_shortcut=True,
            add_to_menu={
                "menu": ApplicationMenus.Source,
                "section": SourceMenuSections.Linting,
            },
            shortcut_widget_context=Qt.ApplicationShortcut,
        )

    @on_plugin_teardown(plugin=Plugins.Editor)
    def on_editor_teardown(self):
        widget = self.get_widget()
        editor = self.get_plugin(Plugins.Editor)

        # Connect to Editor
        widget.sig_edit_goto_requested.disconnect(editor.load)
        editor.sig_editor_focus_changed.disconnect(self._set_filename)

    @on_plugin_teardown(plugin=Plugins.Preferences)
    def on_preferences_teardown(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.deregister_plugin_preferences(self)

    @on_plugin_teardown(plugin=Plugins.Projects)
    def on_projects_teardown(self):
        # Disconnect from projects
        projects = self.get_plugin(Plugins.Projects)
        projects.sig_project_loaded.disconnect(self._set_project_dir)
        projects.sig_project_closed.disconnect(self._unset_project_dir)

    @on_plugin_teardown(plugin=Plugins.Run)
    def on_run_teardown(self):
        run = self.get_plugin(Plugins.Run)
        run.deregister_executor_configuration(
            self, self.executor_configuration)
        run.destroy_run_in_executor_button(
            RunContext.File,
            self.NAME)

    # ---- Private API
    # -------------------------------------------------------------------------
    @Slot()
    def _set_filename(self):
        """
        Set filename without code analysis.
        """
        try:
            editor = self.get_plugin(Plugins.Editor)
            if editor:
                self.get_widget().set_filename(editor.get_current_filename())
        except SpyderAPIError:
            # Editor was deleted
            pass

    def _set_project_dir(self, value):
        widget = self.get_widget()
        widget.set_conf("project_dir", value)

    def _unset_project_dir(self, _unused):
        widget = self.get_widget()
        widget.set_conf("project_dir", None)

    # ---- Public API
    # -------------------------------------------------------------------------
    def change_history_depth(self, value=None):
        """
        Change history maximum number of entries.

        Parameters
        ----------
        value: int or None, optional
            The valur to set  the maximum history depth. If no value is
            provided, an input dialog will be launched. Default is None.
        """
        self.get_widget().change_history_depth(value=value)

    def get_filename(self):
        """
        Get current filename in combobox.
        """
        return self.get_widget().get_filename()

    @run_execute(context=RunContext.File)
    def run_file(
        self,
        input: RunConfiguration,
        conf: ExtendedRunExecutionParameters
    ) -> List[PossibleRunResult]:

        self.switch_to_plugin()
        run_input: FileRun = input['run_input']
        filename = run_input['path']
        self.start_code_analysis(filename)

    @Slot()
    def start_code_analysis(self, filename=None):
        """
        Perform code analysis for given `filename`.

        If `filename` is None default to current filename in combobox.

        If this method is called while still running it will stop the code
        analysis.
        """
        editor = self.get_plugin(Plugins.Editor)
        if editor:
            if self.get_conf("save_before", True) and not editor.save():
                return

        if filename is None or isinstance(filename, bool):
            filename = self.get_widget().get_filename()

        self.switch_to_plugin(force_focus=True)
        self.get_widget().start_code_analysis(filename)

    def stop_code_analysis(self):
        """
        Stop the code analysis process.
        """
        self.get_widget().stop_code_analysis()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2009- Spyder Project Contributors
#
# Distributed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Pythonpath manager plugin.
"""

from qtpy.QtCore import Signal

from spyder.api.plugins import Plugins, SpyderPluginV2
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.translations import _
from spyder.plugins.application.api import ApplicationActions
from spyder.plugins.mainmenu.api import ApplicationMenus, ToolsMenuSections
from spyder.plugins.toolbar.api import ApplicationToolbars, MainToolbarSections
from spyder.plugins.pythonpath.container import (
    PythonpathActions, PythonpathContainer)


class PythonpathManager(SpyderPluginV2):
    """
    Pythonpath manager plugin.
    """

    NAME = "pythonpath_manager"
    REQUIRES = [Plugins.Toolbar, Plugins.MainMenu]
    OPTIONAL = [Plugins.Projects]
    CONTAINER_CLASS = PythonpathContainer
    CONF_SECTION = NAME
    CONF_FILE = False

    sig_pythonpath_changed = Signal(object, bool)
    """
    This signal is emitted when there is a change in the Pythonpath handled by
    Spyder.

    Parameters
    ----------
    new_path_list: list of str
        New list of PYTHONPATH paths.

    prioritize
        Whether to prioritize PYTHONPATH in sys.path
    """

    # ---- SpyderPluginV2 API
    @staticmethod
    def get_name():
        return _("PYTHONPATH manager")

    @staticmethod
    def get_description():
        return _("Manage additional locations to search for Python modules.")

    @classmethod
    def get_icon(cls):
        return cls.create_icon('python')

    def on_initialize(self):
        container = self.get_container()
        container.sig_pythonpath_changed.connect(self.sig_pythonpath_changed)

    @on_plugin_available(plugin=Plugins.MainMenu)
    def on_main_menu_available(self):
        container = self.get_container()
        main_menu = self.get_plugin(Plugins.MainMenu)

        main_menu.add_item_to_application_menu(
            container.path_manager_action,
            menu_id=ApplicationMenus.Tools,
            section=ToolsMenuSections.Managers,
            before=ApplicationActions.SpyderUserEnvVariables,
            before_section=ToolsMenuSections.Preferences,
        )

    @on_plugin_available(plugin=Plugins.Projects)
    def on_projects_available(self):
        projects = self.get_plugin(Plugins.Projects)
        projects.sig_project_loaded.connect(self._on_project_loaded)
        projects.sig_project_closed.connect(self._on_project_closed)

    @on_plugin_available(plugin=Plugins.Toolbar)
    def on_toolbar_available(self):
        container = self.get_container()
        toolbar = self.get_plugin(Plugins.Toolbar)
        toolbar.add_item_to_application_toolbar(
            container.path_manager_action,
            toolbar_id=ApplicationToolbars.Main,
            section=MainToolbarSections.ApplicationSection
        )

    @on_plugin_teardown(plugin=Plugins.MainMenu)
    def on_main_menu_teardown(self):
        main_menu = self.get_plugin(Plugins.MainMenu)
        main_menu.remove_item_from_application_menu(
            PythonpathActions.Manager,
            menu_id=ApplicationMenus.Tools,
        )

    @on_plugin_teardown(plugin=Plugins.Projects)
    def on_projects_teardown(self):
        projects = self.get_plugin(Plugins.Projects)
        projects.sig_project_loaded.disconnect(self._on_project_loaded)
        projects.sig_project_closed.disconnect(self._on_project_closed)

    @on_plugin_teardown(plugin=Plugins.Toolbar)
    def on_toolbar_teardown(self):
        toolbar = self.get_plugin(Plugins.Toolbar)
        toolbar.remove_item_from_application_toolbar(
            PythonpathActions.Manager,
            toolbar_id=ApplicationToolbars.Main
        )

    # ---- Public API
    def get_spyder_pythonpath(self):
        """Get Pythonpath paths handled by Spyder as a list."""
        return self.get_container().get_spyder_pythonpath()

    def show_path_manager(self):
        """Show Path manager dialog."""
        self.get_container().show_path_manager()

    @property
    def path_manager_dialog(self):
        return self.get_container().path_manager_dialog

    # ---- Private API
    def _on_project_loaded(self, path):
        self.get_container().update_active_project_path(path)

    def _on_project_closed(self, path):
        self.get_container().update_active_project_path(None)



# -----------------------------------------------------------------------------
# Copyright (c) 2021- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Spyder API helper decorators.
"""

from __future__ import annotations

# Standard library imports
import functools
import sys
from collections.abc import Callable
from typing import Union

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias  # noqa: ICN003

# Local imports
from spyder.config.types import ConfigurationKey


ConfigurationKeyList: TypeAlias = list[ConfigurationKey]
"""Type alias for a list of string/string tuple keys of Spyder config options.

A :class:`list` of :class:`spyder.config.types.ConfigurationKey`\\s.
"""

ConfigurationKeyOrList: TypeAlias = Union[
    ConfigurationKeyList, ConfigurationKey
]
"""Type alias for either a list of config keys or a single key.

Union of types :class:`ConfigurationKeyList` and
:class:`spyder.config.types.ConfigurationKey`.
"""


def on_conf_change(
    func: Callable | None = None,
    section: str | None = None,
    option: ConfigurationKeyOrList | None = None,
) -> Callable:
    """
    Decorator to handle changing a config option in a given section.

    The methods that use this decorator must have the signature

    .. code-block:: python

        def method(self, value: Any):
            ...

    when observing a single value or the whole section, and

    .. code-block:: python

        def method(self, option: ConfigurationKeyOrList | None, value: Any):
            ...

    when observing multiple values.

    Parameters
    ----------
    func: Callable | None, optional
        Method to decorate, passed automatically when applying the decorator.
    section: str | None, optional
        Name of the configuration section to observe for changes.
        If ``None``, then the ``CONF_SECTION`` attribute of the class
        where the method defined is used.
    option: ConfigurationKeyOrList | None, optional
        Name (:class:`str` / :class:`tuple` of :class:`str`) of the option
        to observe, or a list of names if the method expects updates from
        multiple keys. If ``None``, then all changes to options in the
        specified section are observed.

    Returns
    -------
    func: Callable
        The method passed as ``func`` with the config listener set up.
    """
    if func is None:
        return functools.partial(
            on_conf_change, section=section, option=option
        )

    if option is None:
        # Use special __section identifier to signal that the function
        # observes any change on the section options.
        option = "__section"

    info = []
    if isinstance(option, list):
        info = [(section, opt) for opt in option]
    else:
        info = [(section, option)]

    func._conf_listen = info
    return func



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Status widget for Spyder updates.
"""

# Standard library imports
import logging

# Third party imports
from qtpy.QtCore import Qt, Signal, Slot
from qtpy.QtWidgets import QLabel

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.menus import SpyderMenu
from spyder.api.widgets.status import StatusBarWidget
from spyder.plugins.updatemanager.widgets.update import (
    CHECKING,
    DOWNLOAD_FINISHED,
    DOWNLOADING_INSTALLER,
    UPDATING_UPDATER,
    INSTALL_ON_CLOSE,
    NO_STATUS,
    PENDING
)
from spyder.utils.icon_manager import ima


# Setup logger
logger = logging.getLogger(__name__)


class UpdateManagerStatus(StatusBarWidget):
    """Status bar widget for update manager."""
    ID = 'update_manager_status'
    INTERACT_ON_CLICK = True

    sig_check_update = Signal()
    """Signal to request checking for updates."""

    sig_start_update = Signal()
    """Signal to start the update process."""

    sig_show_progress_dialog = Signal()
    """Signal to show the progress dialog."""

    CUSTOM_WIDGET_CLASS = QLabel

    def __init__(self, parent):

        self.tooltip = ""
        super().__init__(parent)

        # Check for updates action menu
        self.menu = SpyderMenu(self)

        # Set aligment attributes for custom widget to match default label
        # values
        self.custom_widget.setAlignment(Qt.AlignRight | Qt.AlignVCenter)

        # Signals
        self.sig_clicked.connect(self.show_dialog_or_menu)

    def set_value(self, value):
        """Set update manager status."""
        if value == DOWNLOADING_INSTALLER:
            self.tooltip = _(
                "Downloading the update will continue in the background.\n"
                "Click here to show the download dialog again."
            )
            self.custom_widget.show()
            self.show()
        elif value == CHECKING:
            self.tooltip = value
            self.custom_widget.hide()
            self.hide()
        elif value == PENDING:
            self.tooltip = value
            self.custom_widget.hide()
            self.show()
        elif value == UPDATING_UPDATER:
            self.tooltip = value
            self.custom_widget.hide()
            self.show()
        else:
            self.tooltip = ""
            if self.custom_widget:
                self.custom_widget.hide()
            self.hide()

        self.update_tooltip()
        logger.debug(f"Update manager status: {value}")
        super().set_value(value)

    def set_no_status(self):
        """Convenience method to set status to NO_STATUS"""
        self.set_value(NO_STATUS)

    def get_tooltip(self):
        """Reimplementation to get a dynamic tooltip."""
        return self.tooltip

    def get_icon(self):
        return ima.icon('update')

    def set_download_progress(self, percent_progress):
        """Set download progress in status bar"""
        self.custom_widget.setText(f"{percent_progress}%")

    @Slot()
    def show_dialog_or_menu(self):
        """Show download dialog or status bar menu."""
        if self.value in (DOWNLOADING_INSTALLER, UPDATING_UPDATER):
            self.sig_show_progress_dialog.emit()
        elif self.value in (PENDING, DOWNLOAD_FINISHED, INSTALL_ON_CLOSE):
            self.sig_start_update.emit()



# -----------------------------------------------------------------------------
# Copyright (c) 2025- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

from __future__ import annotations

# Standard library imports
import gc
import sys
from typing import TYPE_CHECKING

# Third party imports
import pytest
from _pytest.fixtures import FixtureDef, SubRequest
from qtpy.QtWidgets import QMainWindow

# Local imports
from spyder.api.plugin_registration.registry import PLUGIN_REGISTRY
from spyder.app.cli_options import get_options
from spyder.config.manager import CONF

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Iterable

    from spyder.api.plugin_registration.registry import SpyderPluginClass

__all__ = [
    "main_window_mock",
    "plugins_cls",
    "register_fixture",
]


class MainWindowMock(QMainWindow):
    """:class:`QMainWindow` mock for plugin tests."""

    def __init__(self):
        # This avoids using the cli options passed to pytest
        sys_argv = [sys.argv[0]]
        self._cli_options = get_options(sys_argv)[0]
        super().__init__()
        PLUGIN_REGISTRY.main = self

    def register_plugin(self, plugin_class: type[SpyderPluginClass]):
        plugin = PLUGIN_REGISTRY.register_plugin(self, plugin_class)
        plugin._register()
        return plugin

    @staticmethod
    def unregister_plugin(plugin: SpyderPluginClass):
        assert PLUGIN_REGISTRY.delete_plugin(
            plugin.NAME
        ), f"{plugin.NAME} not deleted"
        plugin._unregister()

    @staticmethod
    def get_plugin(plugin_name, error=False):
        return PLUGIN_REGISTRY.get_plugin(plugin_name)

    @staticmethod
    def is_plugin_available(plugin_name):
        return PLUGIN_REGISTRY.is_plugin_available(plugin_name)


@pytest.fixture(scope="session")
def main_window_mock(qapp):
    """Create a :class:`QMainWindow` mock for plugin tests."""

    window = MainWindowMock()

    try:
        yield window
    finally:
        CONF.reset_manager()
        PLUGIN_REGISTRY.reset()
        del window
        gc.collect()


@pytest.fixture(scope="session")
def plugins_cls() -> (
    Generator[Iterable[tuple[str, type[SpyderPluginClass]]], None, None]
):
    """Fixture that yields plugin classes to be tested.

    Before the yield statement, it will be run at startup.
    After the yield statement, it will be run at teardown.

    Yields
    ------
    list[tuple[str, type[SpyderPluginClass]]]
        A list of fixture names and plugin classes to create a fixture for.

    Raises
    ------
    NotImplementedError
        This fixture must be implemented by the test.
    """
    raise NotImplementedError("This fixture must be implemented by the test.")


@pytest.fixture(scope="session", autouse=True)
def register_fixture(request: SubRequest, plugins_cls) -> None:
    """Dynamically add a fixture for registering plugins."""

    for fixture_name, plugin_cls in plugins_cls:
        # Create a factory function so each fixture gets its own function
        def register_plugin_factory(
            plugin_cls: type[SpyderPluginClass],
        ) -> Callable:
            def register_plugin(main_window_mock):
                plugin = main_window_mock.register_plugin(plugin_cls)
                try:
                    yield plugin
                finally:
                    main_window_mock.unregister_plugin(plugin)

            return register_plugin

        request._fixturemanager._arg2fixturedefs[fixture_name] = [
            FixtureDef(
                argname=fixture_name,
                func=register_plugin_factory(plugin_cls),
                scope="session",
                fixturemanager=request._fixturemanager,
                baseid=request.node.nodeid,
                params=None,
            )
        ]



# -----------------------------------------------------------------------------
# Copyright (c) 2020- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Spyder API auxiliary widgets.
"""

from __future__ import annotations

# Standard library imports
from typing import TYPE_CHECKING

# Third party imports
from qtpy.QtCore import QEvent, QSize, Signal
from qtpy.QtWidgets import QMainWindow, QSizePolicy, QToolBar, QWidget

# Local imports
from spyder.api.exceptions import SpyderAPIError
from spyder.api.widgets import PluginMainWidgetWidgets
from spyder.api.widgets.mixins import SpyderMainWindowMixin
from spyder.utils.stylesheet import APP_STYLESHEET

if TYPE_CHECKING:
    from qtpy.QtGui import QCloseEvent

    import spyder.utils.qthelpers  # For SpyderAction
    from spyder.api.widgets.main_widget import PluginMainWidget


class SpyderWindowWidget(QMainWindow, SpyderMainWindowMixin):
    """
    An undocked window for a :class:`~spyder.api.plugins.SpyderDockablePlugin`.
    """

    # ---- Signals
    # ------------------------------------------------------------------------
    sig_closed: Signal = Signal()
    """Signal emitted when the close event is fired."""

    sig_window_state_changed: Signal = Signal(object)
    """
    Signal is emitted when this window's state has changed.

    For instance, if the window changed between maximized and minimized state.

    Parameters
    ----------
    window_state: Qt.WindowStates
        The new window state that was changed to.
    """

    def __init__(self, widget: PluginMainWidget) -> None:
        """
        Create a window for a :class:`~spyder.api.plugins.SpyderDockablePlugin`.

        Parameters
        ----------
        widget : PluginMainWidget
            The main widget of this window's corresponding
            :class:`~spyder.api.plugins.SpyderDockablePlugin`.

        Returns
        -------
        None
        """
        super().__init__()

        self.widget: PluginMainWidget = widget
        """The main widget of this window's corresponding dockable plugin."""

        self.is_window_widget: bool = True
        """Distinguish these windows from the main Spyder one."""

        # Set interface theme
        self.setStyleSheet(str(APP_STYLESHEET))

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Event handler for when the window is closed.

        Overrides the default :meth:`QWidget.closeEvent` method to emit
        a custom :attr:`sig_closed` signal.
        """
        super().closeEvent(event)
        self.sig_closed.emit()

    def changeEvent(self, event: QEvent) -> None:
        """
        Event handler for when the window state is changed (e.g. minimized).

        Overrides the default :meth:`QWidget.changeEvent` method to emit
        a custom :attr:`sig_window_state_changed` signal.
        """
        if event.type() == QEvent.WindowStateChange:
            self.sig_window_state_changed.emit(self.windowState())
        super().changeEvent(event)


class MainCornerWidget(QToolBar):
    """
    Toolbar widget displayed in the top right hand corner of dockable plugins.

    It is used to display the options (hamburger) menu, progress spinner
    and additional toolbar items to the right of the main toolbar.
    """

    def __init__(self, parent: PluginMainWidget, name: str) -> None:
        """
        Create a new corner widget for a plugin's toolbar.

        Parameters
        ----------
        widget : PluginMainWidget
            The main widget of this window's corresponding
            :class:`~spyder.api.plugins.SpyderDockablePlugin`.
        name : str
            Name of this corner widget, nominally
            :attr:`spyder.api.widgets.PluginMainWidgetWidgets.CornerWidget`.

        Returns
        -------
        None
        """
        super().__init__(parent)
        self._icon_size = QSize(16, 16)
        self.setIconSize(self._icon_size)

        self._widgets = {}
        self._actions = []
        self.setObjectName(name)

        # We add an strut widget here so that there is a spacing
        # between the first item of the corner widget and the last
        # item of the MainWidgetToolbar.
        self._strut = QWidget()
        self._strut.setFixedWidth(0)
        self._strut.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.addWidget(self._strut)

    def add_widget(
        self,
        widget: spyder.utils.qthelpers.SpyderAction | QWidget,
        before: spyder.utils.qthelpers.SpyderAction | QWidget | None = None,
    ) -> None:
        """
        Add a widget to the corner toolbar.

        By default, widgets are added to the left of the last toolbar item.
        Corner widgets provide an options menu button and a spinner so any
        additional widgets will be placed the left of the spinner, if visible
        (unless ``before`` is set).

        Parameters
        ----------
        widget : spyder.utils.qthelpers.SpyderAction | QWidget
            The action or widget to add to the toolbar.
        before : spyder.utils.qthelpers.SpyderAction | QWidget | None, optional
            The action or widget to add ``widget`` before (to the right of).
            If ``None`` (the default), the widget will be added to the left
            of the left-most widget.

        Returns
        -------
        None

        Raises
        ------
        SpyderAPIError
            If either ``widget`` or ``before`` lacks a ``name`` attribute;
            a widget with the same ``name`` as ``widget`` was already added;
            a widget with ``before.name`` has not been added previously; or
            the first widget added is not the options (hamburger) menu widget.
        """
        if not hasattr(widget, "name") or (
            before is not None and not hasattr(before, "name")
        ):
            raise SpyderAPIError(
                f"Widget {widget} or {before} doesn't have a name, which must "
                f"be provided by the attribute `name`"
            )

        if widget.name in self._widgets:
            raise SpyderAPIError(
                'Widget with name "{}" already added. Current names are: {}'
                "".format(widget.name, list(self._widgets.keys()))
            )

        if before is not None and before.name not in self._widgets:
            raise SpyderAPIError(
                f"Widget with name '{before.name}' not in this corner widget"
            )

        if (
            not self._widgets
            and widget.name != PluginMainWidgetWidgets.OptionsToolButton
        ):
            raise SpyderAPIError(
                "The options button must be the first one to be added to the "
                "corner widget of dockable plugins."
            )

        if widget.name == PluginMainWidgetWidgets.OptionsToolButton:
            # This is only necessary for the options button because it's the
            # first one to be added
            action = self.addWidget(widget)
        else:
            if before is not None:
                before_action = self.get_action(before.name)
            else:
                # By default other buttons are added to the left of the last
                # one
                before_action = self._actions[-1]

            # Allow to add either widgets or actions
            if isinstance(widget, QWidget):
                action = self.insertWidget(before_action, widget)
            else:
                action = widget
                self.insertAction(before_action, action)
                widget = self.widgetForAction(action)
                widget.name = action.name

        self._widgets[widget.name] = (widget, action)
        self._actions.append(action)

    def get_widget(self, widget_id: str) -> QWidget | None:
        """
        Return a widget by its unique ID (i.e. its ``name`` attribute).

        Parameters
        ----------
        widget_id : str
            The ``name`` attribute of the widget to return.

        Returns
        -------
        QWidget | None
            The widget object corresponding to ``widget_id``, or ``None``
            if a widget with that ``name`` does not exist.
        """
        if widget_id in self._widgets:
            return self._widgets[widget_id][0]
        return None

    def get_action(
        self, widget_id: str
    ) -> spyder.utils.qthelpers.SpyderAction | None:
        """
        Return an action by its unique ID (i.e. its ``name`` attribute).

        Parameters
        ----------
        widget_id : str
            The ``name`` attribute of the action to return.

        Returns
        -------
        spyder.utils.qthelpers.SpyderAction | None
            The action object corresponding to ``widget_id``, or ``None``
            if an action with that ``name`` does not exist.
        """
        if widget_id in self._widgets:
            return self._widgets[widget_id][1]
        return None



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Spyder IPythonConsole API."""

# Standard library imports
from __future__ import annotations
from typing import TypedDict

# Third-party imports
from typing_extensions import NotRequired  # Available from Python 3.11


class IPythonConsolePyConfiguration(TypedDict):
    """IPythonConsole python execution parameters."""

    # True if the execution is using the current console. False otherwise
    current: bool

    # If True, then the console will start a debugging session if an error
    # occurs. False otherwise.
    post_mortem: bool

    # True if the console is using custom Python arguments. False otherwise.
    python_args_enabled: bool

    # Custom arguments to pass to the console.
    python_args: str

    # If True, then the console will clear all variables before execution.
    # False otherwise.
    clear_namespace: bool

    # If True, then the console will reuse the current namespace. If False,
    # then it will use an empty one.
    console_namespace: bool

    # If not None, then the console will use an alternative run method
    # (e.g. `runfile`, `debugfile` or `debugcell`).
    run_method: NotRequired[str]


class IPythonConsoleWidgetActions:
    # Clients creation
    CreateNewClient = 'new tab'
    CreateCythonClient = 'create cython client'
    CreateSymPyClient = 'create cympy client'
    CreatePyLabClient = 'create pylab client'
    CreateNewClientEnvironment = 'create environment client'

    # Current console actions
    ConnectToKernel = 'connect to kernel'
    Interrupt = 'interrupt kernel'
    Restart = 'Restart kernel'
    Reconnect = 'reconnect kernel'
    ResetNamespace = 'reset namespace'
    ShowEnvironmentVariables = 'Show environment variables'
    ShowSystemPath = 'show system path'
    ToggleElapsedTime = 'toggle elapsed time'

    # Tabs
    RenameTab = 'rename tab'

    # Documentation and help
    IPythonDocumentation = 'ipython documentation'
    ConsoleHelp = 'console help'
    QuickReference = 'quick reference'

    # Navigation
    NextConsole = "switch to next console"
    PreviousConsole = "switch to previous console"


class IPythonConsoleWidgetMenus:
    SpecialConsoles = 'special_consoles_submenu'
    Documentation = 'documentation_submenu'
    EnvironmentConsoles = 'environment_consoles_submenu'
    ClientContextMenu = 'client_context_menu'
    TabsContextMenu = 'tabs_context_menu'


class IPythonConsoleWidgetOptionsMenuSections:
    Edit = 'edit_section'
    View = 'view_section'
    Switch = 'switch_section'


class IPythonConsoleWidgetTabsContextMenuSections:
    Consoles = 'tabs_consoles_section'
    Edit = 'tabs_edit_section'


class ClientContextMenuSections:
    Edit = 'edit'
    Inspect = 'inspect'
    Array = 'array'
    Export = 'export'
    Clear = 'clear'
    Image = 'image'
    SVG = 'svg'
    Quit = 'exit'


class ClientContextMenuActions:
    # Edit section
    Cut = 'cut'
    Copy = 'copy'
    CopyRaw = 'copy_raw'
    Paste = 'paste'
    SelectAll = 'select_all'

    # Inspect section
    InspectObject = 'Inspect current object'

    # Array section
    ArrayInline = 'enter array inline'
    ArrayTable = 'enter array table'

    # Export section
    Export = 'export'
    Print = 'print'

    # Clear section
    ClearConsole = 'Clear shell'
    ClearLine = 'clear line'

    # Image section
    CopyImage = 'copy_image'
    SaveImage = 'save_image'

    # Svg section
    CopySvg = 'copy_svg'
    SaveSvg = 'save_svg'

    # Quit section
    Quit = 'exit'


class IPythonConsoleWidgetCornerWidgets:
    ClearButton = "clear_button"
    InterruptButton = "interrupt_button"
    TimeElapsedLabel = "time_elapsed_label"
    ReconnectButton = "reconnect_button"

class RemoteConsolesMenus:
    RemoteConsoles = "remote_consoles_menu"


class RemoteConsolesMenuSections:
    ManagerSection = "manager_section"
    ConsolesSection = "consoles_section"



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Profiler Plugin.
"""

# Standard library imports
from typing import List

# Third party imports
from packaging.version import parse

# Local imports
from spyder.api.plugins import Plugins, SpyderDockablePlugin
from spyder.api.plugin_registration.decorators import (
    on_plugin_available, on_plugin_teardown)
from spyder.api.shellconnect.mixins import ShellConnectPluginMixin
from spyder.api.translations import _
from spyder.plugins.mainmenu.api import ApplicationMenus, RunMenuSections
from spyder.plugins.profiler.confpage import ProfilerConfigPage
from spyder.plugins.profiler.widgets.main_widget import ProfilerWidget
from spyder.plugins.toolbar.api import ApplicationToolbars
from spyder.plugins.ipythonconsole.api import IPythonConsolePyConfiguration
from spyder.plugins.run.api import (
    ExtendedRunExecutionParameters,
    RunConfiguration,
    RunContext,
    RunExecutor,
    RunResult,
    run_execute,
)
from spyder.plugins.ipythonconsole.widgets.run_conf import IPythonConfigOptions
from spyder.plugins.editor.api.run import CellRun, SelectionRun


# --- Plugin
# ----------------------------------------------------------------------------
class Profiler(SpyderDockablePlugin, ShellConnectPluginMixin, RunExecutor):
    """
    Profiler (after python's profile and pstats).
    """

    NAME = 'profiler'
    REQUIRES = [Plugins.Preferences, Plugins.IPythonConsole, Plugins.Run]
    OPTIONAL = [Plugins.Editor]
    TABIFY = [Plugins.VariableExplorer, Plugins.Help]
    WIDGET_CLASS = ProfilerWidget
    CONF_SECTION = NAME
    CONF_WIDGET_CLASS = ProfilerConfigPage
    CONF_FILE = False

    # ---- SpyderDockablePlugin API
    # -------------------------------------------------------------------------
    @staticmethod
    def get_name():
        return _("Profiler")

    @staticmethod
    def get_description():
        return _("Profile Python files to find execution bottlenecks.")

    @classmethod
    def get_icon(cls):
        return cls.create_icon('profiler')

    def on_initialize(self):

        self.python_editor_run_configuration = {
            'origin': self.NAME,
            'extension': ['py', 'ipy', 'pyw'],
            "contexts": [
                {"name": "File"},
                {"name": "Cell"},
                {"name": "Selection"},
            ],
        }

        self.executor_configuration = [
            {
                'input_extension': 'py',
                'context': {'name': 'File'},
                'output_formats': [],
                'configuration_widget': IPythonConfigOptions,
                'requires_cwd': True,
                'priority': 3
            },
            {
                'input_extension': 'ipy',
                'context': {'name': 'File'},
                'output_formats': [],
                'configuration_widget': IPythonConfigOptions,
                'requires_cwd': True,
                'priority': 10
            },
            {
                'input_extension': 'pyw',
                'context': {'name': 'File'},
                'output_formats': [],
                'configuration_widget': IPythonConfigOptions,
                'requires_cwd': True,
                'priority': 10
            },
            {
                'input_extension': ['py', 'ipy', 'pyw'],
                'context': {'name': 'Cell'},
                'output_formats': [],
                'configuration_widget': None,
                'requires_cwd': True,
                'priority': 10
            },
            {
                'input_extension': ['py', 'ipy', 'pyw'],
                'context': {'name': 'Selection'},
                'output_formats': [],
                'configuration_widget': None,
                'requires_cwd': True,
                'priority': 10
            },
        ]

    @on_plugin_available(plugin=Plugins.Run)
    def on_run_available(self):
        run = self.get_plugin(Plugins.Run)

        # Remove current parameters to recreate them due to the new
        # architecture for profiling introduced in PR spyder-ide/spyder#24794
        if (
            # This is needed when updgrading from Spyder 6.0 to 6.1
            (
                parse(self.old_spyder_conf_version) <= parse("87.3.0")
                and parse(self.spyder_conf_version) > parse("87.3.0")
            )
            # And this when downgrading from Spyder 6.1 to 6.0
            or (
                parse(self.old_spyder_conf_version) > parse("87.3.0")
                and parse(self.spyder_conf_version) <= parse("87.3.0")
            )
        ):
            all_execution_params = self.get_conf(
                "parameters", section="run", default={}
            )
            if self.NAME in all_execution_params:
                all_execution_params.pop(self.NAME)
                self.set_conf(
                    "parameters", all_execution_params, section="run"
                )

        run.register_executor_configuration(self, self.executor_configuration)

        run.create_run_in_executor_button(
            RunContext.File,
            self.NAME,
            text=_("Profile file"),
            tip=_("Profile file"),
            icon=self.create_icon('profiler'),
            shortcut_context=self.NAME,
            register_shortcut=True,
            add_to_menu={
                "menu": ApplicationMenus.Run,
                "section": RunMenuSections.Profile,
            },
            add_to_toolbar=ApplicationToolbars.Profile
        )

        run.create_run_in_executor_button(
            RunContext.Cell,
            self.NAME,
            text=_("Profile cell"),
            tip=_("Profile cell"),
            icon=self.create_icon('profile_cell'),
            shortcut_context=self.NAME,
            register_shortcut=True,
            add_to_menu={
                "menu": ApplicationMenus.Run,
                "section": RunMenuSections.Profile,
            },
            add_to_toolbar=ApplicationToolbars.Profile
        )

        run.create_run_in_executor_button(
            RunContext.Selection,
            self.NAME,
            text=_("Profile current line or selection"),
            tip=_("Profile current line or selection"),
            icon=self.create_icon('profile_selection'),
            shortcut_context=self.NAME,
            register_shortcut=True,
            add_to_menu={
                "menu": ApplicationMenus.Run,
                "section": RunMenuSections.Profile,
            },
            add_to_toolbar=ApplicationToolbars.Profile
        )

    @on_plugin_teardown(plugin=Plugins.Run)
    def on_run_teardown(self):
        run = self.get_plugin(Plugins.Run)
        run.deregister_executor_configuration(
            self, self.executor_configuration
        )

    @on_plugin_available(plugin=Plugins.Editor)
    def on_editor_available(self):
        widget = self.get_widget()
        editor = self.get_plugin(Plugins.Editor)

        editor.add_supported_run_configuration(
            self.python_editor_run_configuration
        )

        widget.sig_edit_goto_requested.connect(editor.load)

    @on_plugin_teardown(plugin=Plugins.Editor)
    def on_editor_teardown(self):
        widget = self.get_widget()
        editor = self.get_plugin(Plugins.Editor)

        editor.remove_supported_run_configuration(
            self.python_editor_run_configuration
        )

        widget.sig_edit_goto_requested.disconnect(editor.load)

    @on_plugin_available(plugin=Plugins.Preferences)
    def on_preferences_available(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.register_plugin_preferences(self)

    @on_plugin_teardown(plugin=Plugins.Preferences)
    def on_preferences_teardown(self):
        preferences = self.get_plugin(Plugins.Preferences)
        preferences.deregister_plugin_preferences(self)

    def on_mainwindow_visible(self):
        # Make plugin visible in case it's not but only once. For most users
        # this will display it in the UI when moving from 6.0 to 6.1
        if not self.get_conf("make_visible", default=False):
            if not self.get_widget().is_visible:
                self.get_widget().toggle_view(True)
            self.set_conf("make_visible", True)

    # ---- For execution
    # -------------------------------------------------------------------------
    @run_execute(context=RunContext.File)
    def profile_file(
        self,
        input: RunConfiguration,
        conf: ExtendedRunExecutionParameters
    ) -> List[RunResult]:

        console = self.get_plugin(Plugins.IPythonConsole)
        if console is None:
            return

        exec_params = conf['params']
        params: IPythonConsolePyConfiguration = exec_params['executor_params']
        params["run_method"] = "profilefile"

        return console.exec_files(input, conf)

    @run_execute(context=RunContext.Cell)
    def profile_cell(
        self,
        input: RunConfiguration,
        conf: ExtendedRunExecutionParameters
    ) -> List[RunResult]:
        console = self.get_plugin(Plugins.IPythonConsole)
        if console is None:
            return

        run_input: CellRun = input['run_input']
        if run_input['copy']:
            code = run_input['cell']
            if not code.strip():
                # Empty cell
                return

            console.run_selection("%%profile\n" + code)
            return

        exec_params = conf['params']
        params: IPythonConsolePyConfiguration = exec_params['executor_params']
        params["run_method"] = "profilecell"

        return console.exec_cell(input, conf)

    @run_execute(context=RunContext.Selection)
    def profile_selection(
        self,
        input: RunConfiguration,
        conf: ExtendedRunExecutionParameters
    ) -> List[RunResult]:

        console = self.get_plugin(Plugins.IPythonConsole)
        if console is None:
            return

        run_input: SelectionRun = input['run_input']
        code = run_input['selection']
        if not code.strip():
            # No selection
            return

        run_input['selection'] = "%%profile\n" + code

        return console.exec_selection(input, conf)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
Editor mixin and utils to manage connection with the LSP
"""

# Standard library imports
import functools
import logging
import random
import re

# Third party imports
from diff_match_patch import diff_match_patch
from qtpy.QtCore import (
    QEventLoop,
    Qt,
    QTimer,
    QThread,
    Signal,
    Slot,
)
from qtpy.QtGui import QColor, QTextCursor
from three_merge import merge

# Local imports
from spyder.config.base import running_under_pytest
from spyder.plugins.completion.api import (
    CompletionRequestTypes,
    TextDocumentSyncKind,
    DiagnosticSeverity,
)
from spyder.plugins.completion.decorators import (
    request,
    handles,
    class_register,
)
from spyder.plugins.editor.panels.utils import (
    merge_folding,
    collect_folding_regions,
)
from spyder.plugins.editor.utils.editor import BlockUserData
from spyder.utils import sourcecode


logger = logging.getLogger(__name__)

# Regexp to detect noqa inline comments.
NOQA_INLINE_REGEXP = re.compile(r"#?noqa", re.IGNORECASE)


def schedule_request(req=None, method=None, requires_response=True):
    """Call function req and then emit its results to the completion server."""
    if req is None:
        return functools.partial(
            schedule_request,
            method=method,
            requires_response=requires_response,
        )

    @functools.wraps(req)
    def wrapper(self, *args, **kwargs):
        params = req(self, *args, **kwargs)
        if params is not None and self.completions_available:
            self._pending_server_requests.append(
                (method, params, requires_response)
            )
            self._server_requests_timer.setInterval(
                self.LSP_REQUESTS_SHORT_DELAY
            )
            self._server_requests_timer.start()

    return wrapper


class LSPHandleError(Exception):
    """Error raised if there is an error handling an LSP response."""


@class_register
class LSPMixin:
    # -- LSP constants
    # Timeouts (in milliseconds) to sychronize symbols and folding after
    # linting results arrive, according to the number of lines in the file.
    SYNC_SYMBOLS_AND_FOLDING_TIMEOUTS = {
        # Lines: Timeout
        500: 600,
        1500: 800,
        2500: 1000,
        6500: 1500,
    }

    # Timeout (in milliseconds) to send pending requests to LSP server
    LSP_REQUESTS_SHORT_DELAY = 50
    LSP_REQUESTS_LONG_DELAY = 300

    # -- LSP signals
    #: Signal emitted when an LSP request is sent to the LSP manager
    sig_perform_completion_request = Signal(str, str, dict)

    #: Signal emitted when a response is received from the completion plugin
    # For now it's only used on tests, but it could be used to track
    # and profile completion diagnostics.
    completions_response_signal = Signal(str, object)

    #: Signal to display object information on the Help plugin
    sig_display_object_info = Signal(str, bool)

    #: Signal only used for tests
    # TODO: Remove it!
    sig_signature_invoked = Signal(dict)

    #: Signal emitted when processing code analysis warnings is finished
    sig_process_code_analysis = Signal()

    #: Signal emitted to tell cloned editors that they need to update their
    # code folding.
    sig_update_code_folding = Signal(tuple)

    # Used to start the status spinner in the editor
    sig_start_operation_in_progress = Signal()

    # Used to start the status spinner in the editor
    sig_stop_operation_in_progress = Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Request symbols and folding after a timeout.
        # See: process_diagnostics
        # Connecting the timeout signal is performed in document_did_open()
        self._timer_sync_symbols_and_folding = QTimer(self)
        self._timer_sync_symbols_and_folding.setSingleShot(True)
        self.blockCountChanged.connect(
            self.set_sync_symbols_and_folding_timeout)

        # LSP requests handling
        # self.textChanged.connect(self._schedule_document_did_change)
        self._pending_server_requests = []
        self._server_requests_timer = QTimer(self)
        self._server_requests_timer.setSingleShot(True)
        self._server_requests_timer.setInterval(self.LSP_REQUESTS_SHORT_DELAY)
        self._server_requests_timer.timeout.connect(
            self._process_server_requests)

        # Code Folding
        self.code_folding = True
        self.update_folding_thread = QThread(None)
        self.update_folding_thread.finished.connect(
            self._finish_update_folding)

        # Autoformat on save
        self.format_on_save = False
        self.format_eventloop = QEventLoop(None)
        self.format_timer = QTimer(self)
        self.__cursor_position_before_format = 0

        # Outline explorer
        self.oe_proxy = None

        # Diagnostics
        self.update_diagnostics_thread = QThread(None)
        self.update_diagnostics_thread.run = self.set_errors
        self.update_diagnostics_thread.finished.connect(
            self.finish_code_analysis)
        self._diagnostics = []

        # Text diffs across versions
        self.differ = diff_match_patch()
        self.previous_text = ''
        self.patch = []
        self.leading_whitespaces = {}

        # Other attributes
        self.filename = None
        self.completions_available = False
        self.text_version = 0
        self.save_include_text = True
        self.open_close_notifications = True
        self.sync_mode = TextDocumentSyncKind.FULL
        self.will_save_notify = False
        self.will_save_until_notify = False
        self.enable_hover = False
        self.auto_completion_characters = []
        self.resolve_completions_enabled = False
        self.signature_completion_characters = []
        self.go_to_definition_enabled = False
        self.find_references_enabled = False
        self.highlight_enabled = False
        self.formatting_enabled = False
        self.range_formatting_enabled = False
        self.document_symbols_enabled = False
        self.formatting_characters = []
        self.completion_args = None
        self.folding_supported = False
        self._folding_info = None
        self.is_cloned = False
        self.operation_in_progress = False
        self.formatting_in_progress = False
        self.symbols_in_sync = False
        self.folding_in_sync = False
        self.pyflakes_linting_enabled = True

    # ---- Helper private methods
    # -------------------------------------------------------------------------
    def _process_server_requests(self):
        """Process server requests."""
        # Check if the document needs to be updated
        if self._document_server_needs_update:
            self.document_did_change()
            self.do_automatic_completions()
            self._document_server_needs_update = False

        # Send pending requests
        for method, params, requires_response in self._pending_server_requests:
            self.emit_request(method, params, requires_response)

        # Clear pending requests
        self._pending_server_requests = []

    # ---- Basic methods
    # -------------------------------------------------------------------------
    @Slot(str, dict)
    def handle_response(self, method, params):
        if method in self.handler_registry:
            handler_name = self.handler_registry[method]
            handler = getattr(self, handler_name)
            handler(params)
            # This signal is only used on tests.
            # It could be used to track and profile LSP diagnostics.
            self.completions_response_signal.emit(method, params)

    def emit_request(self, method, params, requires_response):
        """Send request to LSP manager."""
        params["requires_response"] = requires_response
        params["response_instance"] = self
        self.sig_perform_completion_request.emit(
            self.language.lower(), method, params
        )

    def manage_lsp_handle_errors(self, message):
        """
        Actions to take when we get errors while handling LSP responses.
        """
        # Raise exception so that handle response errors can be reported to
        # Github
        raise LSPHandleError(message)

    # ---- Configuration and start/stop
    # -------------------------------------------------------------------------
    def start_completion_services(self):
        """Start completion services for this instance."""
        self.completions_available = True

        if self.is_cloned:
            additional_msg = "cloned editor"
        else:
            additional_msg = ""
            self.document_did_open()

        logger.debug(
            "Completion services available for {0}: {1}".format(
                additional_msg, self.filename
            )
        )

    def register_completion_capabilities(self, capabilities):
        """
        Register completion server capabilities.

        Parameters
        ----------
        capabilities: dict
            Capabilities supported by a language server.
        """
        sync_options = capabilities["textDocumentSync"]
        save_options = sync_options.get("save", {})
        completion_options = capabilities["completionProvider"]
        signature_options = capabilities["signatureHelpProvider"]
        range_formatting_options = capabilities[
            "documentOnTypeFormattingProvider"
        ]

        self.open_close_notifications = sync_options.get("openClose", False)
        self.sync_mode = sync_options.get("change", TextDocumentSyncKind.NONE)
        self.will_save_notify = sync_options.get("willSave", False)
        self.will_save_until_notify = sync_options.get(
            "willSaveWaitUntil", False
        )
        self.save_include_text = save_options.get("includeText", False)
        self.enable_hover = capabilities["hoverProvider"]
        self.folding_supported = capabilities.get(
            "foldingRangeProvider", False
        )
        self.auto_completion_characters = completion_options[
            "triggerCharacters"
        ]
        self.resolve_completions_enabled = completion_options.get(
            "resolveProvider", False
        )
        self.signature_completion_characters = signature_options[
            "triggerCharacters"
        ] + [
            "="
        ]  # FIXME:
        self.go_to_definition_enabled = capabilities["definitionProvider"]
        self.find_references_enabled = capabilities["referencesProvider"]
        self.highlight_enabled = capabilities["documentHighlightProvider"]
        self.formatting_enabled = capabilities["documentFormattingProvider"]
        self.range_formatting_enabled = capabilities[
            "documentRangeFormattingProvider"
        ]
        self.document_symbols_enabled = capabilities["documentSymbolProvider"]
        self.formatting_characters.append(
            range_formatting_options["firstTriggerCharacter"]
        )
        self.formatting_characters += range_formatting_options.get(
            "moreTriggerCharacter", []
        )

        if self.formatting_enabled:
            self.format_action.setEnabled(True)
            self.sig_refresh_formatting.emit(True)

        self.completions_available = True

    def stop_completion_services(self):
        logger.debug("Stopping completion services for %s" % self.filename)
        self.completions_available = False

    @request(
        method=CompletionRequestTypes.DOCUMENT_DID_OPEN,
        requires_response=False,
    )
    def document_did_open(self):
        """Send textDocument/didOpen request to the server."""

        # We need to be sure that this signal is disconnected before trying to
        # connect it below.
        # Note: It can already be connected when the user requires a server
        # restart or when the server failed to start.
        # Fixes spyder-ide/spyder#20679
        try:
            self._timer_sync_symbols_and_folding.timeout.disconnect()
        except (TypeError, RuntimeError):
            pass

        # The connect is performed here instead of in __init__() because
        # notify_close() may have been called (which disconnects the signal).
        # Qt.UniqueConnection is used to avoid duplicate signal-slot
        # connections (just in case).
        #
        # Note: PyQt5 throws if the signal is not unique (= already connected).
        # It is an error if this happens because as per LSP specification
        # `didOpen` “must not be sent more than once without a corresponding
        # close notification send before”.
        self._timer_sync_symbols_and_folding.timeout.connect(
            self.sync_symbols_and_folding, Qt.UniqueConnection
        )

        cursor = self.textCursor()
        text = self.get_text_with_eol()
        if self.is_ipython():
            # Send valid python text to LSP as it doesn't support IPython
            text = self.ipython_to_python(text)
        params = {
            "file": self.filename,
            "language": self.language,
            "version": self.text_version,
            "text": text,
            "codeeditor": self,
            "offset": cursor.position(),
            "selection_start": cursor.selectionStart(),
            "selection_end": cursor.selectionEnd(),
        }
        return params

    # ---- Symbols
    # -------------------------------------------------------------------------
    @schedule_request(method=CompletionRequestTypes.DOCUMENT_SYMBOL)
    def request_symbols(self):
        """Request document symbols."""
        if not self.document_symbols_enabled:
            return
        if self.oe_proxy is not None:
            self.oe_proxy.emit_request_in_progress()
        params = {"file": self.filename}
        return params

    @handles(CompletionRequestTypes.DOCUMENT_SYMBOL)
    def process_symbols(self, params):
        """Handle symbols response."""
        try:
            symbols = params["params"]
            self._update_classfuncdropdown(symbols)

            if self.oe_proxy is not None:
                self.oe_proxy.update_outline_info(symbols)
        except RuntimeError:
            # This is triggered when a codeeditor instance was removed
            # before the response can be processed.
            return
        except Exception:
            self.manage_lsp_handle_errors("Error when processing symbols")
        finally:
            self.symbols_in_sync = True

    def _update_classfuncdropdown(self, symbols):
        """Update class/function dropdown."""
        symbols = [] if symbols is None else symbols

        if self.classfuncdropdown.isVisible():
            self.classfuncdropdown.update_data(symbols)
        else:
            self.classfuncdropdown.set_data(symbols)

    # ---- Linting and didChange
    # -------------------------------------------------------------------------
    def _schedule_document_did_change(self):
        """Schedule a document update."""
        self._document_server_needs_update = True
        self._server_requests_timer.setInterval(self.LSP_REQUESTS_LONG_DELAY)
        self._server_requests_timer.start()

    @request(
        method=CompletionRequestTypes.DOCUMENT_DID_CHANGE,
        requires_response=False,
    )
    def document_did_change(self):
        """Send textDocument/didChange request to the server."""
        # Cancel formatting
        self.formatting_in_progress = False
        self.symbols_in_sync = False
        self.folding_in_sync = False

        # Don't send request for cloned editors because it's not necessary.
        # The original file should send the request.
        if self.is_cloned:
            return

        # Get text
        text = self.get_text_with_eol()
        if self.is_ipython():
            # Send valid python text to LSP
            text = self.ipython_to_python(text)

        self.text_version += 1

        self.patch = self.differ.patch_make(self.previous_text, text)
        self.previous_text = text
        cursor = self.textCursor()
        params = {
            "file": self.filename,
            "version": self.text_version,
            "text": text,
            "diff": self.patch,
            "offset": cursor.position(),
            "selection_start": cursor.selectionStart(),
            "selection_end": cursor.selectionEnd(),
        }
        return params

    @handles(CompletionRequestTypes.DOCUMENT_PUBLISH_DIAGNOSTICS)
    def process_diagnostics(self, params):
        """Handle linting response."""
        # The LSP spec doesn't require that folding and symbols
        # are treated in the same way as linting, i.e. to be
        # recomputed on didChange, didOpen and didSave. However,
        # we think that's necessary to maintain accurate folding
        # and symbols all the time. Therefore, we decided to call
        # those requests here, but after a certain timeout to
        # avoid performance issues.
        self._timer_sync_symbols_and_folding.start()

        # Process results (runs in a thread)
        self.process_code_analysis(params["params"])

    def set_sync_symbols_and_folding_timeout(self):
        """
        Set timeout to sync symbols and folding according to the file
        size.
        """
        current_lines = self.get_line_count()
        timeout = None

        for lines in self.SYNC_SYMBOLS_AND_FOLDING_TIMEOUTS.keys():
            if (current_lines // lines) == 0:
                timeout = self.SYNC_SYMBOLS_AND_FOLDING_TIMEOUTS[lines]
                break

        if not timeout:
            timeouts = self.SYNC_SYMBOLS_AND_FOLDING_TIMEOUTS.values()
            timeout = list(timeouts)[-1]

        # Add a random number so that several files are not synced at the same
        # time.
        self._timer_sync_symbols_and_folding.setInterval(
            timeout + random.randint(-100, 100)
        )

    def sync_symbols_and_folding(self):
        """
        Synchronize symbols and folding after linting results arrive.
        """
        if not self.folding_in_sync:
            self.request_folding()
        if not self.symbols_in_sync:
            self.request_symbols()

    def process_code_analysis(self, diagnostics):
        """Process code analysis results in a thread."""
        self.cleanup_code_analysis()
        self._diagnostics = diagnostics

        # Process diagnostics in a thread to improve performance.
        self.update_diagnostics_thread.start()

    def cleanup_code_analysis(self):
        """Remove all code analysis markers"""
        self.setUpdatesEnabled(False)
        self.clear_extra_selections("code_analysis_highlight")
        self.clear_extra_selections("code_analysis_underline")
        for data in self.blockuserdata_list():
            data.code_analysis = []

        self.setUpdatesEnabled(True)
        # When the new code analysis results are empty, it is necessary
        # to update manually the scrollflag and linenumber areas (otherwise,
        # the old flags will still be displayed):
        self.sig_flags_changed.emit()
        self.linenumberarea.update()

    def set_errors(self):
        """Set errors and warnings in the line number area."""
        try:
            self._process_code_analysis(underline=False)
        except RuntimeError:
            # This is triggered when a codeeditor instance was removed
            # before the response can be processed.
            return
        except Exception:
            self.manage_lsp_handle_errors("Error when processing linting")

    def underline_errors(self):
        """Underline errors and warnings."""
        try:
            # Clear current selections before painting the new ones.
            # This prevents accumulating them when moving around in or editing
            # the file, which generated a memory leakage and sluggishness
            # after some time.
            self.clear_extra_selections("code_analysis_underline")
            self._process_code_analysis(underline=True)
        except RuntimeError:
            # This is triggered when a codeeditor instance was removed
            # before the response can be processed.
            return
        except Exception:
            self.manage_lsp_handle_errors("Error when processing linting")

    def finish_code_analysis(self):
        """Finish processing code analysis results."""
        self.linenumberarea.update()
        if self.underline_errors_enabled:
            self.underline_errors()
        self.sig_process_code_analysis.emit()
        self.sig_flags_changed.emit()

    def errors_present(self):
        """
        Return True if there are errors or warnings present in the file.
        """
        return bool(len(self._diagnostics))

    def _process_code_analysis(self, underline):
        """
        Process all code analysis results.

        Parameters
        ----------
        underline: bool
            Determines if errors and warnings are going to be set in
            the line number area or underlined. It's better to separate
            these two processes for perfomance reasons. That's because
            setting errors can be done in a thread whereas underlining
            them can't.
        """
        document = self.document()
        if underline:
            first_block, last_block = self.get_buffer_block_numbers()

        for diagnostic in self._diagnostics:
            if self.is_ipython() and (
                diagnostic["message"] == "undefined name 'get_ipython'"
            ):
                # get_ipython is defined in IPython files
                continue
            source = diagnostic.get("source", "")
            msg_range = diagnostic["range"]
            start = msg_range["start"]
            end = msg_range["end"]
            code = diagnostic.get("code", "E")
            message = diagnostic["message"]
            severity = diagnostic.get("severity", DiagnosticSeverity.ERROR)

            block = document.findBlockByNumber(start["line"])
            text = block.text()

            # Skip messages according to certain criteria.
            # This one works for any programming language
            if "analysis:ignore" in text:
                continue

            # This only works for Python and it's only needed with pyflakes.
            if self.language == "Python" and self.pyflakes_linting_enabled:
                if NOQA_INLINE_REGEXP.search(text) is not None:
                    continue

            data = block.userData()
            if not data:
                data = BlockUserData(self)

            if underline:
                block_nb = block.blockNumber()
                if first_block <= block_nb <= last_block:
                    error = severity == DiagnosticSeverity.ERROR
                    color = self.error_color if error else self.warning_color
                    color = QColor(color)
                    color.setAlpha(255)
                    block.color = color

                    data.selection_start = start
                    data.selection_end = end

                    self.highlight_selection(
                        "code_analysis_underline",
                        data._selection(),
                        underline_color=block.color,
                    )
            else:
                # Don't append messages to data for cloned editors to avoid
                # showing them twice or more times on hover.
                # Fixes spyder-ide/spyder#15618
                if not self.is_cloned:
                    data.code_analysis.append(
                        (source, code, severity, message)
                    )
                block.setUserData(data)

    # ---- Completion
    # -------------------------------------------------------------------------
    @schedule_request(method=CompletionRequestTypes.DOCUMENT_COMPLETION)
    def do_completion(self, automatic=False):
        """Trigger completion."""
        cursor = self.textCursor()
        current_word = self.get_current_word(
            completion=True, valid_python_variable=False
        )

        params = {
            "file": self.filename,
            "line": cursor.blockNumber(),
            "column": cursor.columnNumber(),
            "offset": cursor.position(),
            "selection_start": cursor.selectionStart(),
            "selection_end": cursor.selectionEnd(),
            "current_word": current_word,
        }
        self.completion_args = (self.textCursor().position(), automatic)
        return params

    @handles(CompletionRequestTypes.DOCUMENT_COMPLETION)
    def process_completion(self, params):
        """Handle completion response."""
        args = self.completion_args
        if args is None:
            # This should not happen
            return
        self.completion_args = None
        position, automatic = args

        start_cursor = self.textCursor()
        start_cursor.movePosition(QTextCursor.StartOfBlock)
        line_text = self.get_text(start_cursor.position(), "eol")
        leading_whitespace = self.compute_whitespace(line_text)
        indentation_whitespace = " " * leading_whitespace
        eol_char = self.get_line_separator()

        try:
            completions = params["params"]
            completions = (
                []
                if completions is None
                else [
                    completion
                    for completion in completions
                    if completion.get("insertText")
                    or completion.get("textEdit", {}).get("newText")
                ]
            )
            prefix = self.get_current_word(
                completion=True, valid_python_variable=False
            )

            if (
                len(completions) == 1
                and completions[0].get("insertText") == prefix
                and not completions[0].get("textEdit", {}).get("newText")
            ):
                completions.pop()

            replace_end = self.textCursor().position()
            under_cursor = self.get_current_word_and_position(completion=True)
            if under_cursor:
                word, replace_start = under_cursor
            else:
                word = ""
                replace_start = replace_end
            first_letter = ""
            if len(word) > 0:
                first_letter = word[0]

            def sort_key(completion):
                if "textEdit" in completion:
                    text_insertion = completion["textEdit"]["newText"]
                else:
                    text_insertion = completion["insertText"]

                first_insert_letter = text_insertion[0]
                case_mismatch = (
                    first_letter.isupper() and first_insert_letter.islower()
                ) or (first_letter.islower() and first_insert_letter.isupper())

                # False < True, so case matches go first
                return (case_mismatch, completion["sortText"])

            completion_list = sorted(completions, key=sort_key)

            # Allow for textEdit completions to be filtered by Spyder
            # if on-the-fly completions are disabled, only if the
            # textEdit range matches the word under the cursor.
            for completion in completion_list:
                if "textEdit" in completion:
                    c_replace_start = completion["textEdit"]["range"]["start"]
                    c_replace_end = completion["textEdit"]["range"]["end"]

                    if (
                        c_replace_start == replace_start
                        and c_replace_end == replace_end
                    ):
                        insert_text = completion["textEdit"]["newText"]
                        completion["filterText"] = insert_text
                        completion["insertText"] = insert_text
                        del completion["textEdit"]

                if "insertText" in completion:
                    insert_text = completion["insertText"]
                    insert_text_lines = insert_text.splitlines()
                    reindented_text = [insert_text_lines[0]]
                    for insert_line in insert_text_lines[1:]:
                        insert_line = indentation_whitespace + insert_line
                        reindented_text.append(insert_line)
                    reindented_text = eol_char.join(reindented_text)
                    completion["insertText"] = reindented_text

            self.completion_widget.show_list(
                completion_list, position, automatic
            )
        except RuntimeError:
            # This is triggered when a codeeditor instance was removed
            # before the response can be processed.
            return
        except Exception:
            self.manage_lsp_handle_errors("Error when processing completions")

    @schedule_request(method=CompletionRequestTypes.COMPLETION_RESOLVE)
    def resolve_completion_item(self, item):
        return {"file": self.filename, "completion_item": item}

    @handles(CompletionRequestTypes.COMPLETION_RESOLVE)
    def handle_completion_item_resolution(self, response):
        try:
            response = response["params"]

            if not response:
                return

            self.completion_widget.augment_completion_info(response)
        except RuntimeError:
            # This is triggered when a codeeditor instance was removed
            # before the response can be processed.
            return
        except Exception:
            self.manage_lsp_handle_errors(
                "Error when handling completion item resolution"
            )

    # ---- Signature Hints
    # -------------------------------------------------------------------------
    @schedule_request(method=CompletionRequestTypes.DOCUMENT_SIGNATURE)
    def request_signature(self):
        """Ask for signature."""
        line, column = self.get_cursor_line_column()
        offset = self.get_position("cursor")
        params = {
            "file": self.filename,
            "line": line,
            "column": column,
            "offset": offset,
        }
        return params

    @handles(CompletionRequestTypes.DOCUMENT_SIGNATURE)
    def process_signatures(self, params):
        """Handle signature response."""
        try:
            signature_params = params["params"]

            if signature_params is not None:
                self.sig_signature_invoked.emit(signature_params)
                signature_data = signature_params["signatures"]
                documentation = signature_data["documentation"]

                if isinstance(documentation, dict):
                    documentation = documentation["value"]

                # The language server returns encoded text with
                # spaces defined as `\xa0`
                documentation = documentation.replace("\xa0", " ")

                # Enable parsing signature's active parameter if available
                # while allowing to show calltip for signatures without
                # parameters.
                # See spyder-ide/spyder#21660
                parameter = None
                if "activeParameter" in signature_params:
                    parameter_idx = signature_params["activeParameter"]
                    parameters = signature_data["parameters"]
                    if len(parameters) > 0 and parameter_idx < len(parameters):
                        parameter_data = parameters[parameter_idx]
                        parameter = parameter_data["label"]

                signature = signature_data["label"]

                # This method is part of spyder/widgets/mixins
                self.show_calltip(
                    signature=signature,
                    parameter=parameter,
                    language=self.language,
                    documentation=documentation,
                )
        except RuntimeError:
            # This is triggered when a codeeditor instance was removed
            # before the response can be processed.
            return
        except Exception:
            self.manage_lsp_handle_errors("Error when processing signature")

    # ---- Hover/Cursor
    # -------------------------------------------------------------------------
    @schedule_request(method=CompletionRequestTypes.DOCUMENT_CURSOR_EVENT)
    def request_cursor_event(self):
        text = self.get_text_with_eol()
        cursor = self.textCursor()
        params = {
            "file": self.filename,
            "version": self.text_version,
            "text": text,
            "offset": cursor.position(),
            "selection_start": cursor.selectionStart(),
            "selection_end": cursor.selectionEnd(),
        }
        return params

    @schedule_request(method=CompletionRequestTypes.DOCUMENT_HOVER)
    def request_hover(self, line, col, offset, show_hint=True, clicked=True):
        """Request hover information."""
        params = {
            "file": self.filename,
            "line": line,
            "column": col,
            "offset": offset,
        }
        self._show_hint = show_hint
        self._request_hover_clicked = clicked
        return params

    @handles(CompletionRequestTypes.DOCUMENT_HOVER)
    def handle_hover_response(self, contents):
        """Handle hover response."""
        if running_under_pytest():
            from unittest.mock import Mock

            # On some tests this is returning a Mock
            if isinstance(contents, Mock):
                return

        try:
            content = contents["params"]

            # - Don't display hover if there's no content to display.
            # - Prevent spurious errors when a client returns a list.
            if not content or isinstance(content, list):
                return

            self.sig_display_object_info.emit(
                content, self._request_hover_clicked
            )

            if content is not None and self._show_hint and self._last_point:
                # This is located in spyder/widgets/mixins.py
                word = self._last_hover_word

                # Replace non-breaking spaces for real ones.
                content = content.replace("\xa0", " ")

                # Show hover
                self.show_hint(
                    content,
                    inspect_word=word,
                    at_point=self._last_point,
                    vertical_position='top',
                    as_hover=True,
                )

                self._last_point = None
        except RuntimeError:
            # This is triggered when a codeeditor instance was removed
            # before the response can be processed.
            return
        except Exception:
            self.manage_lsp_handle_errors("Error when processing hover")

    # ---- Go To Definition
    # -------------------------------------------------------------------------
    @Slot()
    @schedule_request(method=CompletionRequestTypes.DOCUMENT_DEFINITION)
    def go_to_definition_from_cursor(self, cursor=None):
        """Go to definition from cursor instance (QTextCursor)."""
        if not self.go_to_definition_enabled or self.in_comment_or_string():
            return

        if cursor is None:
            cursor = self.textCursor()

        text = str(cursor.selectedText())

        if len(text) == 0:
            cursor.select(QTextCursor.WordUnderCursor)
            text = str(cursor.selectedText())

        if text is not None:
            line, column = self.get_cursor_line_column()
            params = {"file": self.filename, "line": line, "column": column}
            return params

    @handles(CompletionRequestTypes.DOCUMENT_DEFINITION)
    def handle_go_to_definition(self, position):
        """Handle go to definition response."""
        try:
            position = position["params"]
            if position is not None:
                def_range = position["range"]
                start = def_range["start"]
                if self.filename == position["file"]:
                    self.go_to_line(
                        start["line"] + 1, start["character"], None, word=None
                    )
                else:
                    self.go_to_definition.emit(
                        position["file"], start["line"] + 1, start["character"]
                    )
        except RuntimeError:
            # This is triggered when a codeeditor instance was removed
            # before the response can be processed.
            return
        except Exception:
            self.manage_lsp_handle_errors(
                "Error when processing go to definition"
            )

    # ---- Document/Selection formatting
    # -------------------------------------------------------------------------
    def format_document_or_range(self):
        """Format current document or selected text."""
        if self.has_selected_text() and self.range_formatting_enabled:
            self.format_document_range()
        else:
            self.format_document()

    @schedule_request(method=CompletionRequestTypes.DOCUMENT_FORMATTING)
    def format_document(self):
        """Format current document."""
        self.__cursor_position_before_format = self.textCursor().position()

        if not self.formatting_enabled:
            return
        if self.formatting_in_progress:
            # Already waiting for a formatting
            return

        using_spaces = self.indent_chars != "\t"
        tab_size = (
            len(self.indent_chars)
            if using_spaces
            else self.tab_stop_width_spaces
        )
        params = {
            "file": self.filename,
            "options": {
                "tab_size": tab_size,
                "insert_spaces": using_spaces,
                "trim_trailing_whitespace": self.remove_trailing_spaces,
                "insert_final_new_line": self.add_newline,
                "trim_final_new_lines": self.remove_trailing_newlines,
            },
        }

        # Sets the document into read-only and updates its corresponding
        # tab name to display the filename into parenthesis
        self.setReadOnly(True)
        self.document().setModified(True)
        self.sig_start_operation_in_progress.emit()
        self.operation_in_progress = True
        self.formatting_in_progress = True

        return params

    @schedule_request(method=CompletionRequestTypes.DOCUMENT_RANGE_FORMATTING)
    def format_document_range(self):
        """Format selected text."""
        self.__cursor_position_before_format = self.textCursor().position()

        if not self.range_formatting_enabled or not self.has_selected_text():
            return
        if self.formatting_in_progress:
            # Already waiting for a formatting
            return

        start, end = self.get_selection_start_end()
        start_line, start_col = start
        end_line, end_col = end

        # Remove empty trailing newline from multiline selection
        if end_line > start_line and end_col == 0:
            end_line -= 1

        fmt_range = {
            "start": {"line": start_line, "character": start_col},
            "end": {"line": end_line, "character": end_col},
        }

        using_spaces = self.indent_chars != "\t"
        tab_size = (
            len(self.indent_chars)
            if using_spaces
            else self.tab_stop_width_spaces
        )

        params = {
            "file": self.filename,
            "range": fmt_range,
            "options": {
                "tab_size": tab_size,
                "insert_spaces": using_spaces,
                "trim_trailing_whitespace": self.remove_trailing_spaces,
                "insert_final_new_line": self.add_newline,
                "trim_final_new_lines": self.remove_trailing_newlines,
            },
        }

        # Sets the document into read-only and updates its corresponding
        # tab name to display the filename into parenthesis
        self.setReadOnly(True)
        self.document().setModified(True)
        self.sig_start_operation_in_progress.emit()
        self.operation_in_progress = True
        self.formatting_in_progress = True

        return params

    @handles(CompletionRequestTypes.DOCUMENT_FORMATTING)
    def handle_document_formatting(self, edits):
        """Handle document formatting response."""
        try:
            if self.formatting_in_progress:
                self._apply_document_edits(edits)
        except RuntimeError:
            # This is triggered when a codeeditor instance was removed
            # before the response can be processed.
            return
        except Exception:
            self.manage_lsp_handle_errors(
                "Error when processing document formatting"
            )
        finally:
            # Remove read-only parenthesis and highlight document modification
            self.setReadOnly(False)
            self.document().setModified(False)
            self.document().setModified(True)
            self.sig_stop_operation_in_progress.emit()
            self.operation_in_progress = False
            self.formatting_in_progress = False

    @handles(CompletionRequestTypes.DOCUMENT_RANGE_FORMATTING)
    def handle_document_range_formatting(self, edits):
        """Handle document range formatting response."""
        try:
            if self.formatting_in_progress:
                self._apply_document_edits(edits)
        except RuntimeError:
            # This is triggered when a codeeditor instance was removed
            # before the response can be processed.
            return
        except Exception:
            self.manage_lsp_handle_errors(
                "Error when processing document selection formatting"
            )
        finally:
            # Remove read-only parenthesis and highlight document modification
            self.setReadOnly(False)
            self.document().setModified(False)
            self.document().setModified(True)
            self.sig_stop_operation_in_progress.emit()
            self.operation_in_progress = False
            self.formatting_in_progress = False

    def _apply_document_edits(self, edits):
        """Apply a set of atomic document edits to the current editor text."""
        edits = edits["params"]
        if edits is None:
            return

        # We need to use here toPlainText (which returns text with '\n'
        # for eols) and not get_text_with_eol, so that applying the
        # text edits that come from the LSP in the way implemented below
        # works as expected. That's because we assume eol chars of length
        # one in our algorithm.
        # Fixes spyder-ide/spyder#16180
        text = self.toPlainText()

        text_tokens = list(text)
        merged_text = None
        for edit in edits:
            edit_range = edit["range"]
            repl_text = edit["newText"]
            start, end = edit_range["start"], edit_range["end"]
            start_line, start_col = start["line"], start["character"]
            end_line, end_col = end["line"], end["character"]

            start_pos = self.get_position_line_number(start_line, start_col)
            end_pos = self.get_position_line_number(end_line, end_col)

            # Replace repl_text eols for '\n' to match the ones used in
            # `text`.
            repl_eol = sourcecode.get_eol_chars(repl_text)
            if repl_eol is not None and repl_eol != "\n":
                repl_text = repl_text.replace(repl_eol, "\n")

            text_tokens = list(text_tokens)
            this_edit = list(repl_text)

            if end_line == self.document().blockCount():
                end_pos = self.get_position("eof")
                end_pos += 1

            if (
                end_pos == len(text_tokens)
                and text_tokens[end_pos - 1] == "\n"
            ):
                end_pos += 1

            this_edition = (
                text_tokens[: max(start_pos - 1, 0)]
                + this_edit
                + text_tokens[end_pos - 1:]
            )

            text_edit = "".join(this_edition)
            if merged_text is None:
                merged_text = text_edit
            else:
                merged_text = merge(text_edit, merged_text, text)

        if merged_text is not None:
            # Restore eol chars after applying edits.
            merged_text = merged_text.replace("\n", self.get_line_separator())
            cursor = self.textCursor()

            # Save breakpoints here to restore them after inserting merged_text
            # Fixes spyder-ide/spyder#16549
            if getattr(self, "breakpoints_manager", False):
                breakpoints = self.breakpoints_manager.get_breakpoints()
            else:
                breakpoints = None

            # Begin text insertion
            cursor.beginEditBlock()

            # Select current text
            cursor.movePosition(QTextCursor.Start)
            cursor.movePosition(QTextCursor.End, QTextCursor.KeepAnchor)

            # Insert formatted text in place of the previous one
            cursor.insertText(merged_text)

            # End text insertion
            cursor.endEditBlock()
            
            # Restore breakpoints
            if breakpoints:
                self.breakpoints_manager.set_breakpoints(breakpoints)

            # Restore previous cursor position and center it.
            # Fixes spyder-ide/spyder#19958
            # Use QTextCursor.(position | setPosition) to restore the cursor
            # position to be able to do it with any wrap mode.
            # Fixes spyder-ide/spyder#20852
            if self.__cursor_position_before_format:
                self.moveCursor(QTextCursor.Start)
                cursor = self.textCursor()
                cursor.setPosition(self.__cursor_position_before_format)
                self.setTextCursor(cursor)
                self.centerCursor()

    # ---- Code folding
    # -------------------------------------------------------------------------
    def compute_whitespace(self, line):
        tab_size = self.tab_stop_width_spaces
        whitespace_regex = re.compile(r"(\s+).*")
        whitespace_match = whitespace_regex.match(line)
        total_whitespace = 0
        if whitespace_match is not None:
            whitespace_chars = whitespace_match.group(1)
            whitespace_chars = whitespace_chars.replace("\t", tab_size * " ")
            total_whitespace = len(whitespace_chars)
        return total_whitespace

    def update_whitespace_count(self, line, column):
        self.leading_whitespaces = {}
        lines = str(self.toPlainText()).splitlines()
        for i, text in enumerate(lines):
            total_whitespace = self.compute_whitespace(text)
            self.leading_whitespaces[i] = total_whitespace

    def cleanup_folding(self):
        """Cleanup folding pane."""
        self.folding_panel.folding_regions = {}

    @schedule_request(method=CompletionRequestTypes.DOCUMENT_FOLDING_RANGE)
    def request_folding(self):
        """Request folding."""
        if not self.folding_supported or not self.code_folding:
            return
        params = {"file": self.filename}
        return params

    @handles(CompletionRequestTypes.DOCUMENT_FOLDING_RANGE)
    def handle_folding_range(self, response):
        """Handle folding response."""
        ranges = response["params"]
        if ranges is None:
            return

        # Update folding info in a thread
        self.update_folding_thread.run = functools.partial(
            self._update_folding_info, ranges)
        self.update_folding_thread.start()

    def _update_folding_info(self, ranges):
        """Update folding information with new data from the LSP."""
        try:
            lines = self.toPlainText().splitlines()

            current_tree, root = merge_folding(
                ranges, lines, self.get_line_separator(),
                self.folding_panel.current_tree, self.folding_panel.root
            )

            self._folding_info = (current_tree, root, *collect_folding_regions(root))
        except RuntimeError:
            # This is triggered when a codeeditor instance was removed
            # before the response can be processed.
            return
        except Exception:
            self.manage_lsp_handle_errors("Error when processing folding")

    def highlight_folded_regions(self):
        self.folding_panel.highlight_folded_regions()

    def _finish_update_folding(self):
        """Finish updating code folding."""
        self.sig_update_code_folding.emit(self._folding_info)
        self.apply_code_folding(self._folding_info)

    def apply_code_folding(self, folding_info):
        """Apply code folding info."""
        # Check if we actually have folding info to update before trying to do
        # it.
        # Fixes spyder-ide/spyder#19514
        if folding_info is not None:
            self.folding_panel.update_folding(folding_info)

        self.highlight_folded_regions()

        # Update indent guides, which depend on folding
        if self.indent_guides._enabled:
            line, column = self.get_cursor_line_column()
            self.update_whitespace_count(line, column)

            # This is necessary to repaint guides in cloned editors and the
            # original one after making edits in any one of them.
            # See spyder-ide/spyder#23297
            self.update()

        self.folding_in_sync = True

    # ---- Save/close file
    # -------------------------------------------------------------------------
    @schedule_request(method=CompletionRequestTypes.DOCUMENT_DID_SAVE,
                      requires_response=False)
    def notify_save(self):
        """Send save request."""
        params = {'file': self.filename}
        if self.save_include_text:
            params['text'] = self.get_text_with_eol()
        return params

    @request(method=CompletionRequestTypes.DOCUMENT_DID_CLOSE,
             requires_response=False)
    def notify_close(self):
        """Send close request."""
        self._pending_server_requests = []

        # This is necessary to prevent an error when closing the file.
        # Fixes spyder-ide/spyder#20071
        try:
            self._server_requests_timer.stop()
        except RuntimeError:
            pass

        if self.completions_available:
            # This is necessary to prevent an error in our tests.
            try:
                # Servers can send an empty publishDiagnostics reply to clear
                # diagnostics after they receive a didClose request. Since
                # we also ask for symbols and folding when processing
                # diagnostics, we need to prevent it from happening
                # before sending that request here.
                self._timer_sync_symbols_and_folding.timeout.disconnect()
            except (TypeError, RuntimeError):
                pass

            params = {
                'file': self.filename,
                'codeeditor': self
            }
            return params



# -----------------------------------------------------------------------------
# Copyright (c) 2024- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Class decorator and helpers for running code in async loops within Spyder.

This module provides a class decorator :class:`AsyncDispatcher` to run
coroutines in dedicated async loops, as well as including utilities for
patching loops, managing concurrency tasks, and executing callbacks safely
within Qt applications.
"""

from __future__ import annotations

# Standard library imports
import asyncio
import asyncio.events
import atexit
import collections.abc
import contextlib
import functools
import logging
import os
import sys
import threading
import typing
from asyncio.futures import (
    _chain_future,  # type: ignore[attr-defined] # noqa: PLC2701
)
from asyncio.tasks import (
    _current_tasks,  # type: ignore[attr-defined] # noqa: PLC2701
)
from concurrent.futures import Future
from heapq import heappop

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
    from typing_extensions import TypeAlias
else:
    from typing import ParamSpec  # noqa: ICN003
    from typing import TypeAlias  # noqa: ICN003

# Third party imports
from qtpy.QtCore import QCoreApplication, QEvent, QObject


_logger = logging.getLogger(__name__)

LoopID: TypeAlias = typing.Union[typing.Hashable, asyncio.AbstractEventLoop]
"""Type alias, the union of a loop itself or the ID of one.

Either an existing :class:`asyncio.AbstractEventLoop` or a
:class:`~collections.abc.Hashable` to identify an existing or new one.
"""

_P = ParamSpec("_P")
"""The parameter types of a given async callable."""

_RT = typing.TypeVar("_RT")
"""The return type of a given async callable."""

_T = typing.TypeVar("_T")
"""An arbirary type parameter."""


class AsyncDispatcher(typing.Generic[_RT]):
    """Decorator to run a coroutine in a specific event loop."""

    __rlock = threading.RLock()

    __closed = False
    __running_threads: typing.ClassVar[dict[typing.Hashable, _LoopRunner]] = {}
    _running_tasks: typing.ClassVar[list[Future]] = []

    @typing.overload
    def __init__(
        self: AsyncDispatcher[DispatcherFuture[_T]],
        *,
        loop: LoopID | None = ...,
        early_return: typing.Literal[True] = ...,
        return_awaitable: typing.Literal[False] = ...,
    ) -> None: ...

    @typing.overload
    def __init__(
        self: AsyncDispatcher[typing.Awaitable[_T]],
        *,
        loop: LoopID | None = ...,
        early_return: typing.Literal[True] = ...,
        return_awaitable: typing.Literal[True] = ...,
    ) -> None: ...

    @typing.overload
    def __init__(
        self: AsyncDispatcher[_T],
        *,
        loop: LoopID | None = ...,
        early_return: typing.Literal[False] = ...,
        return_awaitable: typing.Literal[False] = ...,
    ) -> None: ...

    @typing.overload
    def __init__(
        self: AsyncDispatcher[collections.abc.Awaitable[_T]],
        *,
        loop: LoopID | None = ...,
        early_return: typing.Literal[False] = ...,
        return_awaitable: typing.Literal[True] = ...,
    ) -> None: ...

    def __init__(
        self,
        *,
        loop: LoopID | None = None,
        early_return: bool = True,
        return_awaitable: bool = False,
    ) -> None:
        """
        Decorate a coroutine to run in a specific event loop.

        The ``loop`` parameter can be an existing
        :class:`asyncio.AbstractEventLoop` or a
        :class:`~collections.abc.Hashable` to identify an existing or new one
        created by the :class:`!AsyncDispatcher`. If the loop is not running,
        it will be started in a new thread and managed by the
        :class:`!AsyncDispatcher`.

        This instance can be called with the same arguments as the coroutine it
        wraps and will return a :class:`concurrent.futures.Future` object,
        or an awaitable :class:`asyncio.Future` for the current running event
        loop or the result of the coroutine depending on the ``early_return``
        and ``return_awaitable`` parameters.

        Parameters
        ----------
        loop : LoopID | None, optional
            The event loop to be used, by default the current event loop.
        early_return : bool, optional
            Return the coroutine as a :class:`concurrent.futures.Future`
            before it is done. ``True`` by default.
        return_awaitable : bool, optional
            Return the coroutine as an awaitable :class:`asyncio.Future`
            instead of a :class:`concurrent.futures.Future`, independent of
            the value of ``early_return``. ``False`` by default.


        Examples
        --------

        Non-blocking usage (returns a :class:`concurrent.futures.Future`):

        .. code-block:: python

            @AsyncDispatcher()
            async def my_coroutine(...):
                ...

            future = my_coroutine(...)  # Non-blocking call

            result = future.result()  # Blocking call


        Blocking usage (returns the result):

        .. code-block:: python

            @AsyncDispatcher(early_return=False)
            async def my_coroutine(...):
                ...

            result = my_coroutine(...)  # Blocking call

        Coroutine usage (returns an awaitable :class:`asyncio.Future`):

        .. code-block:: python

            @AsyncDispatcher(return_awaitable=True)
            async def my_coroutine(...):
                ...

            result = await my_coroutine(...)  # Wait for the result to be ready
        """
        self._loop = self.get_event_loop(loop)
        self._early_return = early_return
        self._return_awaitable = return_awaitable

    @typing.overload
    def __call__(
        self: AsyncDispatcher[collections.abc.Awaitable[_T]],
        async_func: collections.abc.Callable[
            _P, collections.abc.Awaitable[_T]
        ],
    ) -> collections.abc.Callable[_P, collections.abc.Awaitable[_T]]: ...

    @typing.overload
    def __call__(
        self: AsyncDispatcher[DispatcherFuture[_T]],
        async_func: collections.abc.Callable[
            _P, collections.abc.Awaitable[_T]
        ],
    ) -> collections.abc.Callable[_P, DispatcherFuture[_T]]: ...

    @typing.overload
    def __call__(
        self: AsyncDispatcher[_T],
        async_func: collections.abc.Callable[
            _P, collections.abc.Awaitable[_T]
        ],
    ) -> collections.abc.Callable[_P, _T]: ...

    def __call__(
        self,
        async_func: collections.abc.Callable[
            _P, collections.abc.Awaitable[_T]
        ],
    ) -> collections.abc.Callable[
        _P,
        _T | DispatcherFuture[_T] | collections.abc.Awaitable[_T],
    ]:
        """
        Run the coroutine in the event loop.

        Parameters
        ----------
        *args : tuple
            The positional arguments to be passed to the coroutine.
        **kwargs : dict
            The keyword arguments to be passed to the coroutine.

        Returns
        -------
        concurrent.futures.Future or asyncio.Future or _T

        Raises
        ------
        TypeError
            If ``async_func`` is not a coroutine function.
        """
        if not asyncio.iscoroutinefunction(async_func):
            msg = f"{async_func} is not a coroutine function"
            raise TypeError(msg)

        @functools.wraps(async_func)
        def wrapper(
            *args: _P.args,
            **kwargs: _P.kwargs,
        ) -> _T | DispatcherFuture[_T] | collections.abc.Awaitable[_T]:
            task = run_coroutine_threadsafe(
                async_func(*args, **kwargs),
                loop=self._loop,
            )
            if self._return_awaitable:
                return asyncio.wrap_future(
                    task,
                    loop=asyncio.get_running_loop(),
                )

            if self._early_return:
                AsyncDispatcher._running_tasks.append(task)
                task.add_done_callback(self._callback_task_done)
                return task
            return task.result()

        return wrapper

    @staticmethod
    def _callback_task_done(future: Future):
        AsyncDispatcher._running_tasks.remove(future)

    @classmethod
    def get_event_loop(
        cls,
        loop_id: LoopID | None = None,
    ) -> asyncio.AbstractEventLoop:
        """Get the event loop to run the coroutine.

        If the loop is not running, it will be started in a new thread and
        managed by the :class:`!AsyncDispatcher`.

        Parameters
        ----------
        loop_id : LoopID | None, optional
            The event loop to use, by default the current thread event loop.

            .. note::

                * If a :class:`~collections.abc.Hashable` is provided, it will
                  be used to identify the loop in the
                  :class:`!AsyncDispatcher`.
                * If an event loop is provided, it will be used as the event
                  loop in the :class:`!AsyncDispatcher`.

        Returns
        -------
        asyncio.AbstractEventLoop
            The event loop to be used.
        """
        loop, loop_id = cls._fetch_event_loop(loop_id)

        try:
            if loop.is_running():
                return loop
        except RuntimeError:
            _logger.exception(
                "Failed to check if the loop is running, defaulting to the "
                "current loop.",
            )
            return asyncio.get_event_loop()

        with cls.__rlock:
            # Re-check, perhaps it was created in the meantime...
            if loop_id not in cls.__running_threads:
                cls.__run_loop(loop_id, loop)
                if loop_id is None:
                    asyncio.set_event_loop(loop)

        return loop

    @classmethod
    def _fetch_event_loop(
        cls,
        loop_id: LoopID | None = None,
    ) -> tuple[asyncio.AbstractEventLoop, typing.Hashable | None]:
        """Get the event loop and its hashable id."""
        if loop_id is None:
            with contextlib.suppress(RuntimeError):
                return asyncio.get_running_loop(), None
        elif isinstance(loop_id, asyncio.AbstractEventLoop):
            return loop_id, hash(loop_id)

        running_thread = cls.__running_threads.get(loop_id)
        if running_thread is not None:
            return running_thread.loop, loop_id

        return asyncio.new_event_loop(), loop_id

    @classmethod
    def __run_loop(
        cls,
        loop_id: typing.Hashable,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        if loop_id not in cls.__running_threads:
            with cls.__rlock:
                if loop_id not in cls.__running_threads:
                    _patch_loop_as_reentrant(loop)  # ipykernel compatibility
                    cls.__running_threads[loop_id] = _LoopRunner(loop_id, loop)
                    cls.__running_threads[loop_id].start()

    @staticmethod
    @atexit.register
    def close() -> None:
        """Close the thread pool.

        Returns
        -------
        None
        """
        if AsyncDispatcher.__closed:
            return
        AsyncDispatcher.cancel_all()
        AsyncDispatcher.join()
        AsyncDispatcher.__closed = True

    @classmethod
    def cancel_all(cls) -> None:
        """Cancel all running tasks.

        Returns
        -------
        None
        """
        for task in cls._running_tasks:
            task.cancel()

    @classmethod
    def join(cls, timeout: float | None = None) -> None:
        """Close all running loops and join the threads.

        Parameters
        ----------
        timeout : float | None, optional
            Seconds to wait for loops to exit and their threads to be joined.
            By default, waits indefinitely.

        Returns
        -------
        None
        """
        for loop_id in list(cls.__running_threads.keys()):
            cls._stop_running_loop(loop_id, timeout)

    @classmethod
    def _stop_running_loop(cls, loop_id: LoopID, timeout: float | None = None):
        runner = cls.__running_threads.pop(loop_id, None)

        if runner is None:
            return

        runner.join(timeout)

    @staticmethod
    def QtSlot(
        func: collections.abc.Callable[_P, None],
    ) -> collections.abc.Callable[_P, None]:  # noqa: N802
        """Mark a function to be executed inside the main Qt loop.

        Sets the :attr:`DispatcherFuture.QT_SLOT_ATTRIBUTE` attribute on the
        function to mark it as a slot to be executed in the main Qt loop.

        Parameters
        ----------
        func : Callable
            The function to be marked.

        Returns
        -------
        Callable
            The marked function.
        """
        setattr(func, DispatcherFuture.QT_SLOT_ATTRIBUTE, True)
        return func


class _LoopRunner(threading.Thread):
    """A task runner that runs an asyncio event loop on a background thread."""

    def __init__(
        self,
        loop_id: typing.Hashable,
        loop: asyncio.AbstractEventLoop,
    ):
        super().__init__(daemon=True, name=f"AsyncDispatcher-{loop_id}")
        self.__loop = loop
        self.__loop_stopped = threading.Event()

    @property
    def loop(self):
        return self.__loop

    def run(self):
        asyncio.set_event_loop(self.__loop)
        try:
            self.__loop.run_forever()
        finally:
            self.__loop.close()
            self.__loop_stopped.set()

    def stop(self):
        self.__loop.call_soon_threadsafe(self.__loop.stop)

    def join(self, timeout: float | None = None):
        if not self.__loop_stopped.is_set():
            self.stop()
            self.__loop_stopped.wait(timeout)
        super().join(timeout)


# ruff: noqa: SLF001
# TODO(hlouzada): refactor this function to reduce its complexity
def _patch_loop_as_reentrant(loop):  # noqa: C901, PLR0915
    """
    Patch an event loop in order to make it reentrant.

    This is a simplified version of the 'nest_asyncio'.

    Parameters
    ----------
    loop : asyncio.AbstractEventLoop
        The event loop to be patched.

    Raises
    ------
    TypeError
        If the loop is not an instance of `asyncio.BaseEventLoop`.
    """
    if hasattr(loop, "_nest_patched"):
        """Use same check as asyncio to avoid re-patching."""
        return

    def run_forever(self):
        with manage_run(self), manage_asyncgens(self):
            while True:
                self._run_once()
                if self._stopping:
                    break
        self._stopping = False

    def run_until_complete(self, future):
        with manage_run(self):
            f = asyncio.ensure_future(future, loop=self)
            if f is not future:
                f._log_destroy_pending = False
            while not f.done():
                self._run_once()
                if self._stopping:
                    break
            if not f.done():
                msg = "Event loop stopped before Future completed."
                raise RuntimeError(msg)
            return f.result()

    def _run_once(self):
        """Run handles as they become ready.

        Simplified re-implementation of asyncio's _run_once
        """
        ready = self._ready
        scheduled = self._scheduled
        while scheduled and scheduled[0]._cancelled:
            heappop(scheduled)

        timeout = (
            0
            if ready or self._stopping
            else (
                min(max(scheduled[0]._when - self.time(), 0), 86400)
                if scheduled
                else None
            )
        )
        event_list = self._selector.select(timeout)
        self._process_events(event_list)

        end_time = self.time() + self._clock_resolution
        while scheduled and scheduled[0]._when < end_time:
            handle = heappop(scheduled)
            ready.append(handle)

        for __ in range(len(ready)):
            if not ready:
                break
            handle = ready.popleft()
            if not handle._cancelled:
                # preempt the current task so that that checks in
                # Task.__step do not raise
                curr_task = _current_tasks.pop(self, None)

                try:
                    handle._run()
                finally:
                    # restore the current task
                    if curr_task is not None:
                        _current_tasks[self] = curr_task

        handle = None

    @contextlib.contextmanager
    def manage_run(self):
        """Set up the loop for running."""
        self._check_closed()
        old_thread_id = self._thread_id
        old_running_loop = asyncio.events._get_running_loop()
        try:
            self._thread_id = threading.get_ident()
            asyncio.events._set_running_loop(self)
            self._num_runs_pending += 1
            if self._is_proactorloop and self._self_reading_future is None:
                self.call_soon(self._loop_self_reading)
            yield
        finally:
            self._thread_id = old_thread_id
            asyncio.events._set_running_loop(old_running_loop)
            self._num_runs_pending -= 1
            if (
                self._is_proactorloop
                and self._num_runs_pending == 0
                and self._self_reading_future is not None
            ):
                ov = self._self_reading_future._ov
                self._self_reading_future.cancel()
                if ov is not None:
                    self._proactor._unregister(ov)
                self._self_reading_future = None

    @contextlib.contextmanager
    def manage_asyncgens(self):
        if not hasattr(sys, "get_asyncgen_hooks"):
            # Python version is too old.
            return
        old_agen_hooks = sys.get_asyncgen_hooks()
        try:
            self._set_coroutine_origin_tracking(self._debug)
            if self._asyncgens is not None:
                sys.set_asyncgen_hooks(
                    firstiter=self._asyncgen_firstiter_hook,
                    finalizer=self._asyncgen_finalizer_hook,
                )
            yield
        finally:
            self._set_coroutine_origin_tracking(False)
            if self._asyncgens is not None:
                sys.set_asyncgen_hooks(*old_agen_hooks)

    def _check_running(self):
        """Do not throw exception if loop is already running."""

    if not isinstance(loop, asyncio.BaseEventLoop):
        msg = f"Can't patch loop of type {type(loop)}"
        raise TypeError(msg)

    cls = loop.__class__
    cls.run_forever = run_forever
    cls.run_until_complete = run_until_complete
    cls._run_once = _run_once
    cls._check_running = _check_running
    cls._num_runs_pending = 1 if loop.is_running() else 0
    cls._is_proactorloop = os.name == "nt" and issubclass(
        cls, asyncio.ProactorEventLoop
    )
    cls._nest_patched = True


class _QCallbackEvent(QEvent):
    """Event to execute a callback in the main Qt loop."""

    def __init__(self, func: collections.abc.Callable):
        super().__init__(QEvent.Type.User)
        self.func = func


class _QCallbackExecutor(QObject):
    """Executor to run callbacks in the main Qt loop."""

    def customEvent(self, e: _QCallbackEvent):  # noqa: N802, PLR6301
        e.func()


class DispatcherFuture(Future, typing.Generic[_T]):
    """Represents the result of an asynchronous computation.

    This class is a subclass of :class:`concurrent.futures.Future` that adds
    a :meth:`connect` method to allow attaching callbacks to be executed
    in the main Qt loop.
    """

    QT_SLOT_ATTRIBUTE: str = "__dispatch_qt_slot__"
    """Attribute set on functions to be used as slots in the event loop.

    Normally set by by the :meth:`AsyncDispatcher.QtSlot` decorator.
    """

    _callback_executor = _QCallbackExecutor()

    def result(self, timeout: float | None = None) -> _T:
        """
        Return the result of the call that the future represents.

        Parameters
        ----------
        timeout: float | None, optional
            The number of seconds to wait for the result. If None, then wait
            indefinitely.

        Returns
        -------
        DispatchedFuture@_T
            The result of the call that the future represents.

        Raises
        ------
        asyncio.CancelledError
            If the future was cancelled.

        TimeoutError
            If the future didn't finish executing before the given timeout.

        Exception
            Exception raised by the call that the future represents.
        """  # noqa: DOC502
        return super().result(timeout=timeout)

    def connect(
        self, fn: collections.abc.Callable[[DispatcherFuture[_T]], None]
    ) -> None:
        """Attaches a callable that will be called when the future finishes.

        The callable will be called by a thread in the same process in which
        it was added if the it was not marked with
        :attr:`DispatcherFuture.QT_SLOT_ATTRIBUTE`.

        If the future has already been cancelld or completed then the callable
        will be executed immediately. These callables are called in the order
        that they were added.

        Parameters
        ----------
        fn: Callable
            A callable that will be executed with this future as its argument
            once the future completes or is cancelled.

        Returns
        -------
        None
        """
        if getattr(fn, self.QT_SLOT_ATTRIBUTE, False):

            def callback(future: DispatcherFuture[_T]):
                e = _QCallbackEvent(lambda: fn(future))
                QCoreApplication.postEvent(self._callback_executor, e)

            self.add_done_callback(callback)  # type: ignore[arg-type]
        else:
            self.add_done_callback(fn)  # type: ignore[arg-type]


def run_coroutine_threadsafe(
    coro: collections.abc.Coroutine[_T, None, _RT],
    loop: asyncio.AbstractEventLoop,
) -> DispatcherFuture[_RT]:
    """Submit a coroutine object to run in the given event loop.

    Arguments
    ---------
    coro: Coroutine
        The coroutine object to be submitted.
    loop: asyncio.AbstractEventLoop
        The event loop in which to run the coroutine.

    Returns
    -------
    DispatcherFuture
        A future object representing the result of the coroutine.

    Raises
    ------
    TypeError
        If the object is not a coroutine.
    """
    if not asyncio.iscoroutine(coro):
        msg = "A coroutine object is required"
        raise TypeError(msg)
    future = DispatcherFuture()

    def callback():
        try:
            _chain_future(asyncio.ensure_future(coro, loop=loop), future)
        except (SystemExit, KeyboardInterrupt):
            raise
        except BaseException as exc:
            if future.set_running_or_notify_cancel():
                future.set_exception(exc)
            raise

    loop.call_soon_threadsafe(callback)
    return future



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Editor config page."""

from qtpy.QtWidgets import (
    QButtonGroup,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
)

from spyder.api.config.decorators import on_conf_change
from spyder.api.config.mixins import SpyderConfigurationObserver
from spyder.api.preferences import PluginConfigPage
from spyder.api.translations import _
from spyder.config.manager import CONF
from spyder.plugins.editor.widgets.mouse_shortcuts import MouseShortcutEditor


NUMPYDOC = "https://numpydoc.readthedocs.io/en/latest/format.html"
GOOGLEDOC = (
    "https://sphinxcontrib-napoleon.readthedocs.io/en/latest/"
    "example_google.html"
)
SPHINXDOC = (
    "https://sphinx-rtd-tutorial.readthedocs.io/en/latest/docstrings.html"
)
DOCSTRING_SHORTCUT = CONF.get('shortcuts', 'editor/docstring')


class EditorConfigPage(PluginConfigPage, SpyderConfigurationObserver):

    def __init__(self, plugin, parent):
        PluginConfigPage.__init__(self, plugin, parent)
        SpyderConfigurationObserver.__init__(self)

        self.removetrail_box = None
        self.add_newline_box = None
        self.remove_trail_newline_box = None
        self.tabwidth_spin = None
        self.indent_char_box = None

        self.apply_callback = self.set_indent_chars

    def setup_page(self):
        newcb = self.create_checkbox

        # ---- Display tab
        # -- Interface group
        interface_group = QGroupBox(_("Interface"))
        showtabbar_box = newcb(
            _("Show tab bar"),
            'show_tab_bar',
            tip=_(
                "If hidden, the file switcher, Outline pane and Ctrl-Tab\n"
                "can still be used to navigate between open files."
            ),
        )
        show_filename_box = newcb(
            _("Show full file path above editor"),
            'show_filename_toolbar',
            tip=_(
                "If hidden, the full file path is still shown when hovering "
                "over its tab."
            ),
        )
        showclassfuncdropdown_box = newcb(
            _("Show class/function selector"),
            'show_class_func_dropdown',
            tip=_(
                "For quick browsing and switching between classes/functions "
                "in a file."
            ),
        )
        scroll_past_end_box = newcb(
            _("Allow scrolling past file end"), 'scroll_past_end'
        )

        interface_layout = QVBoxLayout()
        interface_layout.addWidget(showtabbar_box)
        interface_layout.addWidget(show_filename_box)
        interface_layout.addWidget(showclassfuncdropdown_box)
        interface_layout.addWidget(scroll_past_end_box)
        interface_group.setLayout(interface_layout)

        # -- Helpers group
        helpers_group = QGroupBox(_("Helpers"))
        showindentguides_box = newcb(_("Show indent guides"), 'indent_guides')
        showcodefolding_box = newcb(
            _("Show code folding"),
            'code_folding',
            tip=_("Allow collapsing and uncollapsing code by indent level"),
        )
        linenumbers_box = newcb(_("Show line numbers"), 'line_numbers')
        breakpoints_box = newcb(
            _("Show debugger breakpoints"),
            'editor_debugger_panel',
            section='debugger',
        )
        todolist_box = newcb(
            _("Show code annotations"),
            'todo_list',
            tip=_(
                "Display a marker to the left of line numbers when the "
                "following annotations appear at the beginning of a comment: "
                "<tt>TODO, FIXME, XXX, HINT, TIP, @todo, HACK, BUG, OPTIMIZE, "
                "!!!, ???</tt> (and their lowercase variants)"
            ),
        )

        helpers_layout = QVBoxLayout()
        helpers_layout.addWidget(showindentguides_box)
        helpers_layout.addWidget(showcodefolding_box)
        helpers_layout.addWidget(linenumbers_box)
        helpers_layout.addWidget(breakpoints_box)
        helpers_layout.addWidget(todolist_box)
        helpers_group.setLayout(helpers_layout)

        # -- Highlight group
        highlight_group = QGroupBox(_("Highlight"))
        currentline_box = newcb(
            _("Highlight current line"), 'highlight_current_line'
        )
        currentcell_box = newcb(
            _("Highlight current cell"), 'highlight_current_cell'
        )
        occurrence_box = newcb(
            _("Highlight occurrences of selected text after"),
            'occurrence_highlighting',
        )
        occurrence_spin = self.create_spinbox(
            "",
            _(" milliseconds"),
            'occurrence_highlighting/timeout',
            min_=100,  # 0.1 seconds
            max_=60_000,  # 1 minute
            step=100,  # 0.1 seconds
        )

        occurrence_box.checkbox.toggled.connect(
            occurrence_spin.spinbox.setEnabled
        )
        occurrence_box.checkbox.toggled.connect(
            occurrence_spin.slabel.setEnabled
        )
        occurrence_spin.spinbox.setEnabled(
            self.get_option('occurrence_highlighting')
        )
        occurrence_spin.slabel.setEnabled(
            self.get_option('occurrence_highlighting')
        )

        occurrence_glayout = QGridLayout()
        occurrence_glayout.addWidget(occurrence_box, 0, 0)
        occurrence_glayout.addWidget(occurrence_spin.spinbox, 0, 1)
        occurrence_glayout.addWidget(occurrence_spin.slabel, 0, 2)

        occurrence_layout = QHBoxLayout()
        occurrence_layout.addLayout(occurrence_glayout)
        occurrence_layout.addStretch(1)

        highlight_layout = QVBoxLayout()
        highlight_layout.addWidget(currentline_box)
        highlight_layout.addWidget(currentcell_box)
        highlight_layout.addLayout(occurrence_layout)
        highlight_group.setLayout(highlight_layout)

        # ---- Source code tab
        # -- Automatic changes group
        automatic_group = QGroupBox(_("Automatic changes"))
        closepar_box = newcb(
            _("Automatically insert closing parentheses, brackets and braces"),
            'close_parentheses',
            tip=_("Insert a matching closing bracket when typing an open one"),
        )
        close_quotes_box = newcb(
            _("Automatically insert closing quotes"),
            'close_quotes',
            tip=_("Insert a matching closing quote when typing an open one"),
        )
        add_colons_box = newcb(
            _("Automatically insert colons after 'for', 'if', 'def', etc"),
            'add_colons',
        )
        autounindent_box = newcb(
            _("Automatically un-indent 'else', 'elif', etc"),
            'auto_unindent',
            tip=_(
                "Un-indent further block-level keywords "
                "when added inside an 'if', etc block"
            ),
        )

        automatic_layout = QVBoxLayout()
        automatic_layout.addWidget(closepar_box)
        automatic_layout.addWidget(close_quotes_box)
        automatic_layout.addWidget(add_colons_box)
        automatic_layout.addWidget(autounindent_box)
        automatic_group.setLayout(automatic_layout)

        # -- Trailing whitespace group
        whitespace_group = QGroupBox(_("Trailing whitespace"))
        self.removetrail_box = newcb(
            _("Strip all trailing spaces on save"),
            'always_remove_trailing_spaces',
            default=False,
        )
        strip_mode_box = newcb(
            _("Strip trailing spaces on changed lines"),
            'strip_trailing_spaces_on_modify',
            default=True,
            tip=_(
                "If enabled, modified lines of code (excluding strings)\n"
                "will have trailing whitespace stripped when leaving them.\n"
                "If disabled, only whitespace added by Spyder will be "
                "stripped."
            ),
        )
        self.add_newline_box = newcb(
            _("Automatically add missing end-of-file newline on save"),
            'add_newline',
            default=False,
            tip=_(
                "If enabled, a trailing newline character (line break) will "
                "automatically be appended to the end of the file\n"
                "if the file does not already end with one, "
                "to conform to standard text file conventions."
            ),
        )
        self.remove_trail_newline_box = newcb(
            _("Strip blank lines at end of file on save"),
            'always_remove_trailing_newlines',
            default=False,
            tip=_(
                "Any extra newlines at the end of the file besides the first "
                "one will be stripped."
            ),
        )

        # Disable the fix-on-save options if autoformatting is enabled
        format_on_save = CONF.get(
            'completions',
            ('provider_configuration', 'lsp', 'values', 'format_on_save'),
            False,
        )
        self.on_format_save_state(format_on_save)

        whitespace_layout = QVBoxLayout()
        whitespace_layout.addWidget(self.removetrail_box)
        whitespace_layout.addWidget(strip_mode_box)
        whitespace_layout.addWidget(self.add_newline_box)
        whitespace_layout.addWidget(self.remove_trail_newline_box)
        whitespace_group.setLayout(whitespace_layout)

        # -- Identation group
        indentation_group = QGroupBox(_("Indentation"))
        self.tabwidth_spin = self.create_spinbox(
            _("Tab width:"),
            _("spaces"),
            "tab_stop_width_spaces",
            default=4,
            min_=1,
            max_=16,
            step=1,
        )
        self.indent_char_box = newcb(
            _("Indent with spaces instead of tabs"),
            'indent_with_spaces',
            default=True,
            tip=_(
                "When pressing Tab, insert the configured number of spaces "
                "instead of the tab character"
            ),
        )
        ibackspace_box = newcb(
            _("Intelligent backspace"),
            'intelligent_backspace',
            tip=_(
                "Make the backspace key automatically remove the number of "
                "indentation characters set above."
            ),
            default=True,
        )
        tab_mode_box = newcb(
            _("Tab always indents"),
            'tab_always_indent',
            default=False,
            tip=_(
                "If enabled, pressing Tab will always add an indent,\n"
                "even when the cursor is not at the beginning of a line.\n"
                "Code completion can still be triggered using the shortcut "
                "Ctrl+Space."
            ),
        )

        indent_tab_grid_layout = QGridLayout()
        indent_tab_grid_layout.addWidget(self.tabwidth_spin.plabel, 0, 0)
        indent_tab_grid_layout.addWidget(self.tabwidth_spin.spinbox, 0, 1)
        indent_tab_grid_layout.addWidget(self.tabwidth_spin.slabel, 0, 2)

        indent_tab_layout = QHBoxLayout()
        indent_tab_layout.addLayout(indent_tab_grid_layout)
        indent_tab_layout.addStretch(1)

        indentation_layout = QVBoxLayout()
        indentation_layout.addLayout(indent_tab_layout)
        indentation_layout.addWidget(self.indent_char_box)
        indentation_layout.addWidget(ibackspace_box)
        indentation_layout.addWidget(tab_mode_box)
        indentation_group.setLayout(indentation_layout)

        # -- EOL group
        eol_group = QGroupBox(_("End-of-line characters"))
        fix_eol_box = newcb(
            _("Fix mixed end-of-lines automatically and show warning dialog"),
            'check_eol_chars',
            default=True,
            tip=_(
                "When opening a file containing mixed end-of-line characters\n"
                "(which may raise syntax errors in the console on Windows),\n"
                "Spyder will convert them automatically if enabled."
            ),
        )
        convert_eol_on_save_box = newcb(
            _("Convert end-of-line characters to the following on save:"),
            'convert_eol_on_save',
            default=False,
        )
        eol_combo_choices = (
            ("LF (Linux/macOS)", 'LF'),
            ("CRLF (Windows)", 'CRLF'),
            (_("CR (legacy Mac)"), 'CR'),
        )
        convert_eol_on_save_combo = self.create_combobox(
            "",
            eol_combo_choices,
            'convert_eol_on_save_to',
        )

        convert_eol_on_save_box.checkbox.toggled.connect(
            convert_eol_on_save_combo.setEnabled
        )
        convert_eol_on_save_combo.setEnabled(
            self.get_option('convert_eol_on_save')
        )

        eol_on_save_layout = QHBoxLayout()
        eol_on_save_layout.addWidget(convert_eol_on_save_box)
        eol_on_save_layout.addWidget(convert_eol_on_save_combo)

        eol_layout = QVBoxLayout()
        eol_layout.addWidget(fix_eol_box)
        eol_layout.addLayout(eol_on_save_layout)
        eol_group.setLayout(eol_layout)

        # ---- Advanced tab
        # -- Template group
        template_group = QGroupBox(_("Template"))
        template_button = self.create_button(
            text=_("Edit new file template"),
            callback=self.plugin.edit_template,
            set_modified_on_click=True,
        )

        template_layout = QVBoxLayout()
        template_layout.addSpacing(3)
        template_layout.addWidget(template_button)
        template_group.setLayout(template_layout)

        # -- Autosave group
        autosave_group = QGroupBox(_("Autosave"))
        autosave_checkbox = newcb(
            _("Automatically save a backup copy of unsaved files"),
            'autosave_enabled',
            tip=_(
                "If Spyder quits unexpectedly, it will offer to recover "
                "them on next launch"
            ),
        )
        autosave_spinbox = self.create_spinbox(
            _("Autosave interval: "),
            _("seconds"),
            'autosave_interval',
            min_=1,
            max_=3600,
        )

        autosave_checkbox.checkbox.toggled.connect(autosave_spinbox.setEnabled)

        autosave_layout = QVBoxLayout()
        autosave_layout.addWidget(autosave_checkbox)
        autosave_layout.addWidget(autosave_spinbox)
        autosave_group.setLayout(autosave_layout)

        # -- Docstring group
        docstring_group = QGroupBox(_("Docstring style"))
        numpy_url = "<a href='{}'>Numpy</a>".format(NUMPYDOC)
        googledoc_url = "<a href='{}'>Google</a>".format(GOOGLEDOC)
        sphinx_url = "<a href='{}'>Sphinx</a>".format(SPHINXDOC)
        docstring_label = QLabel(
            _(
                "Select the style of docstrings ({numpy}, {google} or {sphinx}) "
                "to generate when pressing <kbd>{shortcut}</kbd> after a "
                "function, method or class declaration"
            ).format(
                numpy=numpy_url,
                google=googledoc_url,
                sphinx=sphinx_url,
                shortcut=DOCSTRING_SHORTCUT,
            ),
        )
        docstring_label.setOpenExternalLinks(True)
        docstring_label.setWordWrap(True)
        docstring_combo_choices = (
            ("Numpy", 'Numpydoc'),
            ("Google", 'Googledoc'),
            ("Sphinx", 'Sphinxdoc'),
        )
        docstring_combo = self.create_combobox(
            _("Style:"),
            docstring_combo_choices,
            'docstring_type',
        )

        docstring_layout = QVBoxLayout()
        docstring_layout.addWidget(docstring_label)
        docstring_layout.addWidget(docstring_combo)
        docstring_group.setLayout(docstring_layout)

        # -- Multi-cursor group
        multicursor_group = QGroupBox(_("Multi-cursor"))
        multicursor_box = newcb(
            _("Enable multi-cursor support"),
            'multicursor_support',
            tip=_(
                "Allows adding additional cursors and columns of cursors "
                "for simultaneous editing"
            ),
        )

        multicursor_layout = QVBoxLayout()
        multicursor_layout.addWidget(multicursor_box)
        multicursor_group.setLayout(multicursor_layout)

        # -- Multi-cursor paste group
        multicursor_paste_group = QGroupBox(_("Multi-cursor paste behavior"))
        multicursor_paste_bg = QButtonGroup(multicursor_paste_group)
        entire_clip_radio = self.create_radiobutton(
            _("Always paste the entire clipboard for each cursor"),
            "multicursor_paste/always_full",
            button_group=multicursor_paste_bg,
        )
        conditional_spread_radio = self.create_radiobutton(
            _(
                "Paste one line per cursor if the number of lines and cursors "
                "match"
            ),
            "multicursor_paste/conditional_spread",
            button_group=multicursor_paste_bg,
        )
        always_spread_radio = self.create_radiobutton(
            _(
                "Always paste one line per cursor if there is more than one "
                "line in the clipboard"
            ),
            "multicursor_paste/always_spread",
            button_group=multicursor_paste_bg,
        )

        multicursor_box.checkbox.toggled.connect(
            multicursor_paste_group.setEnabled
        )
        multicursor_paste_group.setEnabled(
            self.get_option("multicursor_support")
        )

        multicursor_paste_layout = QVBoxLayout()
        multicursor_paste_layout.addWidget(entire_clip_radio)
        multicursor_paste_layout.addWidget(conditional_spread_radio)
        multicursor_paste_layout.addWidget(always_spread_radio)
        multicursor_paste_group.setLayout(multicursor_paste_layout)

        # -- Mouse shortcuts group
        mouse_shortcuts_group = QGroupBox(_("Mouse shortcuts"))
        mouse_shortcuts_button = self.create_button(
            lambda: MouseShortcutEditor(self).exec_(),
            _("Edit mouse shortcut modifiers"),
        )

        mouse_shortcuts_layout = QVBoxLayout()
        mouse_shortcuts_layout.addWidget(mouse_shortcuts_button)
        mouse_shortcuts_group.setLayout(mouse_shortcuts_layout)

        # --- Tabs ---
        self.create_tab(
            _("Display"),
            [
                interface_group,
                helpers_group,
                highlight_group,
            ],
        )

        self.create_tab(
            _("Source code"),
            [
                automatic_group,
                whitespace_group,
                indentation_group,
                eol_group,
            ],
        )

        self.create_tab(
            _("Advanced"),
            [
                template_group,
                autosave_group,
                docstring_group,
                multicursor_group,
                multicursor_paste_group,
                mouse_shortcuts_group,
            ],
        )

    @on_conf_change(
        option=('provider_configuration', 'lsp', 'values', 'format_on_save'),
        section='completions',
    )
    def on_format_save_state(self, value):
        """
        Change options following the `format_on_save` completion option.

        Parameters
        ----------
        value : bool
            If the completion `format_on_save` option is enabled or disabled.

        Returns
        -------
        None.

        """
        options = [
            self.removetrail_box,
            self.add_newline_box,
            self.remove_trail_newline_box,
        ]
        for option in options:
            if option:
                if value:
                    option.setToolTip(
                        _(
                            "This option is disabled since the "
                            "<i>Autoformat files on save</i> option is active."
                        )
                    )
                else:
                    option.setToolTip("")
                option.setDisabled(value)

    def set_indent_chars(self):
        """Set the indent_chars config option per the two sub-options."""
        if (
            "tab_stop_width_spaces" not in self.changed_options
            and "indent_with_spaces" not in self.changed_options
        ):
            return

        indent_tab = not self.indent_char_box.checkbox.isChecked()
        indent_width = (
            1 if indent_tab else int(self.tabwidth_spin.spinbox.value())
        )
        indent_char = "\t" if indent_tab else " "
        indent_chars = "*{}*".format(indent_char * indent_width)

        self.set_option("indent_chars", indent_chars)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Array Builder Widget."""

# TODO:
# - Set font based on caller? editor console? and adjust size of widget
# - Fix positioning
# - Use the same font as editor/console?
# - Generalize separators
# - Generalize API for registering new array builders

# Standard library imports
import re

# Third party imports
import qstylizer.style
from qtpy.QtCore import QEvent, QPoint, Qt
from qtpy.QtGui import QKeyEvent
from qtpy.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QDialog,
    QHBoxLayout,
    QLineEdit,
    QTableWidget,
    QTableWidgetItem,
    QToolButton,
    QToolTip,
    QVBoxLayout,
)

# Local imports
from spyder.api.translations import _
from spyder.config.base import running_under_pytest
from spyder.utils.icon_manager import ima
from spyder.utils.palette import SpyderPalette
from spyder.utils.stylesheet import AppStyle
from spyder.widgets.helperwidgets import TipWidget

# Constants
SHORTCUT_TABLE = "Ctrl+M"
SHORTCUT_INLINE = "Ctrl+Alt+M"


class ArrayBuilderType:
    LANGUAGE = None
    ELEMENT_SEPARATOR = None
    ROW_SEPARATOR = None
    BRACES = None
    EXTRA_VALUES = None
    ARRAY_PREFIX = None

    def check_values(self):
        pass


class ArrayBuilderPython(ArrayBuilderType):
    ELEMENT_SEPARATOR = ', '
    ROW_SEPARATOR = ';'
    BRACES = '], ['
    EXTRA_VALUES = {
        'np.nan': ['nan', 'NAN', 'NaN', 'Na', 'NA', 'na'],
        'np.inf': ['inf', 'INF'],
    }
    ARRAY_PREFIX = 'np.array([['


_REGISTERED_ARRAY_BUILDERS = {
    'python': ArrayBuilderPython,
}


class ArrayInline(QLineEdit):

    def __init__(self, parent, options=None):
        super().__init__(parent)
        self._parent = parent
        self._options = options

    def keyPressEvent(self, event):
        """Override Qt method."""
        if event.key() in [Qt.Key_Enter, Qt.Key_Return]:
            self._parent.process_text()
            if self._parent.is_valid():
                self._parent.keyPressEvent(event)
        else:
            super().keyPressEvent(event)

    # To catch the Tab key event
    def event(self, event):
        """
        Override Qt method.

        This is needed to be able to intercept the Tab key press event.
        """
        if event.type() == QEvent.KeyPress:
            if (event.key() == Qt.Key_Tab or event.key() == Qt.Key_Space):
                text = self.text()
                cursor = self.cursorPosition()

                # Fix to include in "undo/redo" history
                if cursor != 0 and text[cursor-1] == ' ':
                    text = (text[:cursor-1] + self._options.ROW_SEPARATOR
                            + ' ' + text[cursor:])
                else:
                    text = text[:cursor] + ' ' + text[cursor:]
                self.setCursorPosition(cursor)
                self.setText(text)
                self.setCursorPosition(cursor + 1)

                return False

        return super().event(event)


class ArrayTable(QTableWidget):

    def __init__(self, parent, options=None):
        super().__init__(parent)
        self._parent = parent
        self._options = options

        # Default number of rows and columns
        self.setRowCount(2)
        self.setColumnCount(2)

        # Select the first cell and send an Enter to it so users can start
        # editing the array after it's displayed.
        # Idea taken from https://stackoverflow.com/a/32205884/438386
        if not running_under_pytest():
            self.setCurrentCell(0, 0)
            event = QKeyEvent(QEvent.KeyPress, Qt.Key_Enter, Qt.NoModifier)
            QApplication.postEvent(self, event)

        # Set headers
        self.reset_headers()

        # Signals
        self.cellChanged.connect(self.cell_changed)

    def keyPressEvent(self, event):
        shift = event.modifiers() & Qt.ShiftModifier

        if event.key() in [Qt.Key_Enter, Qt.Key_Return] and not shift:
            # To edit a cell when pressing Enter.
            # From https://stackoverflow.com/a/70166617/438386
            if self.state() != QAbstractItemView.EditingState:
                self.edit(self.currentIndex())

        super().keyPressEvent(event)

    def cell_changed(self, row, col):
        item = self.item(row, col)
        value = None

        if item:
            rows = self.rowCount()
            cols = self.columnCount()
            value = item.text()

        if value:
            if row == rows - 1:
                self.setRowCount(rows + 1)
            if col == cols - 1:
                self.setColumnCount(cols + 1)
        self.reset_headers()

    def reset_headers(self):
        """Update the column and row numbering in the headers."""
        rows = self.rowCount()
        cols = self.columnCount()

        for r in range(rows):
            self.setVerticalHeaderItem(r, QTableWidgetItem(str(r)))
        for c in range(cols):
            self.setHorizontalHeaderItem(c, QTableWidgetItem(str(c)))
            self.setColumnWidth(c, 40)

    def text(self):
        """Return the entered array in a parseable form."""
        text = []
        rows = self.rowCount()
        cols = self.columnCount()

        # handle empty table case
        if rows == 2 and cols == 2:
            item = self.item(0, 0)
            if item is None:
                return ''

        for r in range(rows - 1):
            for c in range(cols - 1):
                item = self.item(r, c)
                if item is not None:
                    value = item.text()
                else:
                    value = '0'

                if not value.strip():
                    value = '0'

                text.append(' ')
                text.append(value)
            text.append(self._options.ROW_SEPARATOR)

        return ''.join(text[:-1])  # Remove the final uneeded `;`


class ArrayBuilderDialog(QDialog):

    def __init__(self, parent=None, inline=True, offset=0, force_float=False,
                 language='python'):
        super().__init__(parent=parent)
        self._language = language
        self._options = _REGISTERED_ARRAY_BUILDERS.get('python', None)
        self._parent = parent
        self._text = None
        self._valid = None
        self._offset = offset

        # TODO: add this as an option in the General Preferences?
        self._force_float = force_float

        help_inline = _("""
           <b>Numpy array helper</b><br><br>
           * Type an array using the syntax: <tt>1 2;3 4</tt><br>
           * Use two spaces or tabs to generate a <b>;</b>.<br>
           * You can press <b>Shift+Enter</b> when you are done.
           """
        )

        help_table = _("""
           <b>Numpy array helper</b><br><br>
           * Introduce an array in the table.<br>
           * Use <b>Tab</b> to move between cells or introduce additional
             rows and columns.<br>
           * Use two tabs at the end of a row to move to the next one.<br>
           * You can press <b>Shift+Enter</b> when you are done.
           """
        )

        # Widgets
        button_ok = QToolButton()
        button_ok.setIcon(ima.icon("DialogApplyButton"))
        button_ok.setToolTip(_("Introduce contents"))
        button_ok.clicked.connect(self.accept)

        button_close = QToolButton()
        button_close.setIcon(ima.icon("DialogCloseButton"))
        button_close.setToolTip(_("Cancel"))
        button_close.clicked.connect(self.reject)

        buttons_css = qstylizer.style.StyleSheet()
        buttons_css.QToolButton.setValues(
            height="16px",
            width="16px",
            borderRadius=SpyderPalette.SIZE_BORDER_RADIUS
        )
        for button in [button_ok, button_close]:
            button.setStyleSheet(buttons_css.toString())

        button_help = TipWidget(
            tip_text=help_inline if inline else help_table,
            icon=ima.icon('info_tip'),
            hover_icon=ima.icon('info_tip_hover')
        )

        if inline:
            self._text = ArrayInline(self, options=self._options)
            self._widget = self._text
        else:
            self._table = ArrayTable(self, options=self._options)
            self._widget = self._table

        # Style
        css = qstylizer.style.StyleSheet()
        css.QDialog.setValues(
            margin="0px",
            border=f"1px solid {SpyderPalette.COLOR_BACKGROUND_6}",
            borderRadius=SpyderPalette.SIZE_BORDER_RADIUS,
        )
        self.setStyleSheet(css.toString())

        # widget setup
        self.setWindowFlags(Qt.Window | Qt.Dialog | Qt.FramelessWindowHint)
        self.setModal(True)
        self._widget.setMinimumWidth(200)

        # layout
        if inline:
            buttons_layout = QHBoxLayout()
            buttons_layout.addWidget(button_help, alignment=Qt.AlignVCenter)
            buttons_layout.addWidget(button_ok, alignment=Qt.AlignVCenter)
            buttons_layout.addWidget(button_close, alignment=Qt.AlignVCenter)
        else:
            buttons_layout = QVBoxLayout()
            buttons_layout.addWidget(button_ok)
            buttons_layout.addSpacing(3)
            buttons_layout.addWidget(button_close)
            buttons_layout.addStretch()
            buttons_layout.addWidget(button_help)

        layout = QHBoxLayout()
        layout.setContentsMargins(
            3 * AppStyle.MarginSize,
            3 * AppStyle.MarginSize,
            2 * AppStyle.MarginSize + 1,
            3 * AppStyle.MarginSize
        )
        layout.addWidget(self._widget)
        layout.addLayout(buttons_layout)
        self.setLayout(layout)

        self._widget.setFocus()

    def keyPressEvent(self, event):
        """Override Qt method."""
        QToolTip.hideText()
        shift = event.modifiers() & Qt.ShiftModifier

        if event.key() in [Qt.Key_Enter, Qt.Key_Return] and shift:
            self.accept()
        else:
            super().keyPressEvent(event)

    def event(self, event):
        """
        Override Qt method.

        Useful when in line edit mode.
        """
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Tab:
            return False

        return super().event(event)

    def accept(self):
        self.process_text()
        super().accept()

    def process_text(self):
        """
        Construct the text based on the entered content in the widget.
        """
        prefix = self._options.ARRAY_PREFIX

        suffix = ']])'
        values = self._widget.text().strip()

        if values != '':
            # cleans repeated spaces
            exp = r'(\s*)' + self._options.ROW_SEPARATOR + r'(\s*)'
            values = re.sub(exp, self._options.ROW_SEPARATOR, values)
            values = re.sub(r"\s+", " ", values)
            values = re.sub(r"]$", "", values)
            values = re.sub(r"^\[", "", values)
            values = re.sub(self._options.ROW_SEPARATOR + r'*$', '', values)

            # replaces spaces by commas
            values = values.replace(' ',  self._options.ELEMENT_SEPARATOR)

            # iterate to find number of rows and columns
            new_values = []
            rows = values.split(self._options.ROW_SEPARATOR)
            nrows = len(rows)
            ncols = []
            for row in rows:
                new_row = []
                elements = row.split(self._options.ELEMENT_SEPARATOR)
                ncols.append(len(elements))
                for e in elements:
                    num = e

                    # replaces not defined values
                    for key, values in self._options.EXTRA_VALUES.items():
                        if num in values:
                            num = key

                    # Convert numbers to floating point
                    if self._force_float:
                        try:
                            num = str(float(e))
                        except:
                            pass
                    new_row.append(num)
                new_values.append(
                    self._options.ELEMENT_SEPARATOR.join(new_row))
            new_values = self._options.ROW_SEPARATOR.join(new_values)
            values = new_values

            # Check validity
            if len(set(ncols)) == 1:
                self._valid = True
            else:
                self._valid = False

            # Single rows are parsed as 1D arrays/matrices
            if nrows == 1:
                prefix = prefix[:-1]
                suffix = suffix.replace("]])", "])")

            # Fix offset
            offset = self._offset
            braces = self._options.BRACES.replace(
                ' ',
                '\n' + ' '*(offset + len(prefix) - 1))
            values = values.replace(self._options.ROW_SEPARATOR,  braces)
            text = "{0}{1}{2}".format(prefix, values, suffix)

            self._text = text
        else:
            self._text = ''

        self.update_warning()

    def update_warning(self):
        """
        Updates the icon and tip based on the validity of the array content.
        """
        if not self.is_valid():
            tip = _('Array dimensions are not valid')
            QToolTip.showText(
                self._widget.mapToGlobal(QPoint(3, 18)), tip, self
            )

    def is_valid(self):
        """Return if the current array state is valid."""
        return self._valid

    def text(self):
        """Return the parsed array/matrix text."""
        return self._text

    @property
    def array_widget(self):
        """Return the array builder widget."""
        return self._widget


def test():  # pragma: no cover
    from spyder.utils.qthelpers import qapplication
    from spyder.utils.stylesheet import APP_STYLESHEET

    app = qapplication()
    app.setStyleSheet(str(APP_STYLESHEET))

    dlg_table = ArrayBuilderDialog(None, inline=False)
    dlg_inline = ArrayBuilderDialog(None, inline=True)
    dlg_table.show()
    dlg_inline.show()
    app.exec_()


if __name__ == "__main__":  # pragma: no cover
    test()



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

# Third party imports
from qtpy.QtCore import QSize, Signal, Slot
from qtpy.QtWidgets import QDialog, QDialogButtonBox, QHBoxLayout, QPushButton
from superqt.utils import qdebounced

# Local imports
from spyder.api.translations import _
from spyder.api.widgets.dialogs import SpyderDialogButtonBox
from spyder.config.base import load_lang_conf
from spyder.config.manager import CONF
from spyder.utils.icon_manager import ima
from spyder.utils.stylesheet import MAC, WIN
from spyder.widgets.sidebardialog import SidebarDialog


class ConfigDialog(SidebarDialog):
    """Preferences dialog."""

    # Signals
    check_settings = Signal()
    sig_size_changed = Signal(QSize)
    sig_reset_preferences_requested = Signal()

    # Constants
    TITLE = _("Preferences")
    MIN_WIDTH = 940 if MAC else (875 if WIN else 920)
    MIN_HEIGHT = 700 if MAC else (660 if WIN else 670)

    def __init__(self, parent=None):
        self.ICON = ima.icon('configure')
        SidebarDialog.__init__(self, parent)

        # Attributes
        self.main = parent

        # Ensures that the config is present on spyder first run
        CONF.set('main', 'interface_language', load_lang_conf())

    # ---- Public API
    # -------------------------------------------------------------------------
    def get_index_by_name(self, name):
        """Return page index by CONF_SECTION name."""
        for idx in range(self.pages_widget.count()):
            page = self.get_page(idx)

            # This is the case for separators
            if page is None:
                continue

            section = page.plugin.NAME
            if section == name:
                return idx
        else:
            return None

    def check_all_settings(self):
        """
        This method is called to check all configuration page settings after
        configuration dialog has been shown.
        """
        self.check_settings.emit()

    # ---- SidebarDialog API
    # -------------------------------------------------------------------------
    def button_clicked(self, button):
        if button is self.apply_btn:
            # Apply button was clicked
            configpage = self.get_page()
            if not configpage.is_valid():
                return
            configpage.apply_changes()

    def current_page_changed(self, index):
        widget = self.get_page(index)
        self.apply_btn.setVisible(widget.apply_callback is not None)
        self.apply_btn.setEnabled(widget.is_modified)

    def add_page(self, page, initialize=False):
        # Signals
        self.check_settings.connect(page.check_settings)
        page.apply_button_enabled.connect(self.apply_btn.setEnabled)
        super().add_page(page, initialize=initialize)

    def create_buttons(self):
        bbox = SpyderDialogButtonBox(
            QDialogButtonBox.Ok
            | QDialogButtonBox.Apply
            | QDialogButtonBox.Cancel
        )
        self.apply_btn = bbox.button(QDialogButtonBox.Apply)

        # This is needed for our tests
        self.ok_btn = bbox.button(QDialogButtonBox.Ok)

        button_reset = QPushButton(_('Reset to defaults'))
        button_reset.clicked.connect(self.sig_reset_preferences_requested)
        bbox.addButton(button_reset, QDialogButtonBox.ResetRole)

        layout = QHBoxLayout()
        layout.addWidget(bbox)

        return bbox, layout

    # ---- Qt methods
    # -------------------------------------------------------------------------
    @Slot()
    def accept(self):
        for index in range(self.pages_widget.count()):
            configpage = self.get_page(index)

            # This can be the case for separators, which doesn't have a config
            # page.
            if configpage is None:
                continue

            if not configpage.is_valid():
                return

            configpage.apply_changes()

        QDialog.accept(self)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        self._on_resize()

    # ---- Private API
    # -------------------------------------------------------------------------
    @qdebounced(timeout=40)
    def _on_resize(self):
        """
        We name this method differently from SidebarDialog._on_resize_event
        because we want to debounce this as well.
        """
        self.sig_size_changed.emit(self.size())



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""Language Server Protocol message boxes."""

# Third party imports
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QMessageBox

# Local imports
from spyder.api.translations import _
from spyder.widgets.helperwidgets import MessageCheckBox


class ServerDisabledMessageBox(MessageCheckBox):
    sig_restart_spyder = Signal()

    def __init__(self, parent, warn_str, set_conf):
        super().__init__(icon=QMessageBox.Warning, parent=parent)
        self.set_conf = set_conf

        self.setWindowTitle(_("Warning"))
        self.set_checkbox_text(_("Don't show again"))
        self.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
        self.setDefaultButton(QMessageBox.No)
        self.set_checked(False)
        self.set_check_visible(True)
        self.setText(warn_str)

    def exec_(self):
        answer = super().exec_()
        self.set_conf('show_lsp_down_warning', not self.is_checked())
        if answer == QMessageBox.Yes:
            self.sig_restart_spyder.emit()

    @classmethod
    def instance(cls, warn_str, set_conf):
        def wrapper(parent):
            return cls(parent, warn_str, set_conf)
        return wrapper



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
# (see spyder/__init__.py for details)

"""
IPython Console restart dialog for preferences.
"""

# Third party imports
import qstylizer.style
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QButtonGroup,
    QDialog,
    QLabel,
    QPushButton,
    QRadioButton,
    QVBoxLayout,
)

# Local imports
from spyder.api.translations import _
from spyder.utils.stylesheet import AppStyle, MAC


class ConsoleRestartDialog(QDialog):
    """
    Dialog to apply preferences that need a restart of the console kernel.
    """

    # Constants for actions when Preferences require a kernel restart
    NO_RESTART = 1
    RESTART_CURRENT = 2
    RESTART_ALL = 3

    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowFlags(Qt.Dialog | Qt.WindowTitleHint)
        self._parent = parent
        self.setWindowTitle(_("Restart kernels"))

        self._action = self.NO_RESTART
        self._action_string = {
            self.NO_RESTART: _("Keep existing kernels"),
            self.RESTART_CURRENT: _("Restart current kernel"),
            self.RESTART_ALL: _("Restart all kernels")
        }

        # Initial option values
        self.restart_all = False
        self.restart_current = False
        self.no_restart = True

        # Dialog widgets
        # Text
        self._text_label = QLabel(
            _("By default, some IPython console preferences will be "
              "applied to new consoles only. To apply preferences to "
              "existing consoles, select from the options below.<br><br>"
              "Please note: applying changes to running consoles will force"
              " a kernel restart and all current work will be lost."),
            self)
        self._text_label.setWordWrap(True)
        self._text_label.setFixedWidth(450)

        # Radio buttons
        self._no_restart_btn = QRadioButton(
            _("Don't apply changes to existing consoles"), self
        )
        self._restart_current_btn = QRadioButton(
            _("Apply to current console and restart kernel"), self
        )
        self._restart_all_btn = QRadioButton(
            _("Apply to all existing consoles and restart all kernels"), self
        )

        self._radio_group = QButtonGroup(self)
        self._radio_group.setExclusive(True)
        self._radio_group.addButton(
            self._no_restart_btn, id=self.NO_RESTART
        )
        self._radio_group.addButton(
            self._restart_current_btn, id=self.RESTART_CURRENT
        )
        self._radio_group.addButton(
            self._restart_all_btn, id=self.RESTART_ALL
        )

        self._action_button = QPushButton(
            self._action_string[self.NO_RESTART], parent=self
        )

        # Dialog Layout
        layout = QVBoxLayout(self)
        layout.addWidget(self._text_label)
        layout.addSpacing(2 * AppStyle.MarginSize)
        layout.addWidget(self._no_restart_btn)
        layout.addWidget(self._restart_current_btn)
        layout.addWidget(self._restart_all_btn)
        layout.addSpacing(
            2 * AppStyle.MarginSize if MAC else 3 * AppStyle.MarginSize
        )
        layout.addWidget(self._action_button, 0, Qt.AlignRight)
        layout.setContentsMargins(*((7 * AppStyle.MarginSize,) * 4))
        self.setLayout(layout)

        # Signals
        self._no_restart_btn.setChecked(True)
        self._radio_group.buttonToggled.connect(
            self.update_action_button_text)
        self._action_button.clicked.connect(self.accept)

        # Stylesheet
        self._set_stylesheet()

    def update_action_button_text(self, radiobutton, is_checked):
        """
        Update action button text.

        Takes into account the given radio button to update the text.
        """
        radiobutton_id = self._radio_group.id(radiobutton)
        if is_checked:
            text = self._action_string[radiobutton_id]
            self._radio_group.buttonToggled.disconnect(
                self.update_action_button_text)

            self._no_restart_btn.setChecked(False)
            self._restart_current_btn.setChecked(False)
            self._restart_all_btn.setChecked(False)
            radiobutton.setChecked(True)

            self._radio_group.buttonToggled.connect(
                self.update_action_button_text)
        else:
            text = self._action_string[self.NO_RESTART]
        self._action_button.setText(text)

    def get_action_value(self):
        """
        Return tuple indicating True or False for the available actions.
        """
        self.restart_current = self._restart_current_btn.isChecked()
        self.restart_all = self._restart_all_btn.isChecked()
        self.no_restart = not any([self.restart_all, self.restart_current])
        return self.restart_all, self.restart_current, self.no_restart

    def _set_stylesheet(self):
        radiobuttons_css = qstylizer.style.StyleSheet()
        radiobuttons_css.setValues(
            marginLeft=f"{5 * AppStyle.MarginSize}px"
        )
        for label in [
            self._no_restart_btn,
            self._restart_current_btn,
            self._restart_all_btn,
        ]:
            label.setStyleSheet(radiobuttons_css.toString())

    def exec_(self):
        # Give focus to the last selected button
        if self.restart_all:
            self._restart_all_btn.setFocus()
        elif self.restart_current:
            self._restart_current_btn.setFocus()
        else:
            self._no_restart_btn.setFocus()

        super().exec_()



# -----------------------------------------------------------------------------
# Copyright (c) 2025- Spyder Project Contributors
#
# Released under the terms of the MIT License
# (see LICENSE.txt in the project root directory for details)
# -----------------------------------------------------------------------------

"""
Global registry of Spyder plugins and decorators for setup and teardown.
"""
